@@ -1,5 +1,6 @@
 import cProfile
 import inspect
+import math
 import os.path
 import pstats
 import random
@@ -27,6 +28,7 @@ from fakeparser import Parser
 
 from reader import make_reader
 from reader._app import create_app, get_reader
+from reader._config import make_reader_config
 
 
 def get_params(fn):
@@ -101,7 +103,7 @@ def make_reader_with_entries(path, num_entries, num_feeds=NUM_FEEDS, text=False)
 
 
 def make_test_client(path):
-    app = create_app(path)
+    app = create_app(make_reader_config({'reader': {'url': path}}))
     client = app.test_client()
     with app.app_context():
         get_reader()
@@ -114,8 +116,14 @@ def setup_db():
         yield os.path.join(tmpdir, 'db.sqlite')
 
 
+EXISTING_DB_PATH = None
+
+
 @contextmanager
 def setup_db_with_entries(num_entries):
+    if EXISTING_DB_PATH:
+        yield EXISTING_DB_PATH
+        return
     with setup_db() as path:
         make_reader_with_entries(path, num_entries).update_feeds()
         yield path
@@ -262,6 +270,9 @@ time_update_feed_old_fallback = inject(reader=setup_reader_feed_old_fallback)(
 
 @contextmanager
 def setup_reader_with_text_entries(num_entries):
+    if EXISTING_DB_PATH:
+        yield make_reader(EXISTING_DB_PATH)
+        return
     with setup_db() as path:
         reader = make_reader_with_entries(path, num_entries, text=True)
         reader.update_feeds()
@@ -270,6 +281,9 @@ def setup_reader_with_text_entries(num_entries):
 
 @contextmanager
 def setup_reader_with_search_and_some_read_entries(num_entries):
+    if EXISTING_DB_PATH:
+        yield make_reader(EXISTING_DB_PATH)
+        return
     with setup_reader_with_text_entries(num_entries) as reader:
         reader.enable_search()
         reader.update_search()
@@ -283,17 +297,29 @@ SEARCH_ENTRIES_QUERY = 'porta justo scelerisque dignissim convallis primis lacus
 
 
 @inject(reader=setup_reader_with_search_and_some_read_entries)
-def time_search_entries_all(reader):
+def time_search_entries_relevant_all(reader):
     for _ in reader.search_entries(SEARCH_ENTRIES_QUERY):
         pass
 
 
 @inject(reader=setup_reader_with_search_and_some_read_entries)
-def time_search_entries_read(reader):
+def time_search_entries_relevant_read(reader):
     for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, read=True):
         pass
 
 
+@inject(reader=setup_reader_with_search_and_some_read_entries)
+def time_search_entries_recent_all(reader):
+    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, sort='recent'):
+        pass
+
+
+@inject(reader=setup_reader_with_search_and_some_read_entries)
+def time_search_entries_recent_read(reader):
+    for _ in reader.search_entries(SEARCH_ENTRIES_QUERY, sort='recent', read=True):
+        pass
+
+
 @inject(reader=setup_reader_with_text_entries)
 def time_update_search(reader):
     # Sadly time() doesn't allow running the setup for every repeat,
@@ -311,7 +337,7 @@ TIMINGS = OrderedDict(
 )
 TIMINGS_PARAMS_LIST = [(2 ** i,) for i in range(5, 12)]
 TIMINGS_NUMBER = 4
-PROFILE_PARAMS = (2 ** 11,)
+PROFILE_PARAMS = TIMINGS_PARAMS_LIST[-1]
 PARAM_IDS = ('num_entries',)
 
 
@@ -340,7 +366,14 @@ def make_row_fmt(extra, names, num_fmt='.3f'):
 @click.argument('which', nargs=-1)
 @click.option('-n', '--number', type=int, default=TIMINGS_NUMBER, show_default=True)
 @click.option('-r', '--repeat', type=int, show_default=True)
-def time(which, number, repeat):
+@click.option(
+    '--db',
+    type=click.Path(exists=True, dir_okay=False),
+    help="Use an existing database instead of generating ones of varying sizes. "
+    "WARNING: Benchmarks that mutate the database might not make sense.",
+)
+@click.option('--query', help="Query for search_entries() timings; useful with --db.")
+def time(which, number, repeat, db, query):
     if not which:
         which = ['*']
 
@@ -367,6 +400,17 @@ def time(which, number, repeat):
     header = make_header(extra, names)
     row_fmt = make_row_fmt(extra, names)
 
+    if db:
+        # HACK: use an existing DB
+        # (for which num_entries can't vary, since it exists already)
+        global EXISTING_DB_PATH
+        EXISTING_DB_PATH = db
+        global TIMINGS_PARAMS_LIST
+        TIMINGS_PARAMS_LIST = [(0,)]
+    if query:
+        global SEARCH_ENTRIES_QUERY
+        SEARCH_ENTRIES_QUERY = query
+
     def get_results():
         for params in TIMINGS_PARAMS_LIST:
             times = []
@@ -387,6 +431,13 @@ def time(which, number, repeat):
             print(row_fmt.format(*prefix, *params, *map(stat, results)))
 
 
+def fancy_division(a, b):
+    try:
+        return a / b
+    except ZeroDivisionError:
+        return math.copysign(float('inf'), a)
+
+
 @cli.command()
 @click.argument('before', type=click.File())
 @click.argument('after', type=click.File())
@@ -400,11 +451,11 @@ def time(which, number, repeat):
 def diff(before, after, format):
     if format == 'percent-decrease':
         row_num_fmt = '.1%'
-        func = lambda b, a: 1 - float(a) / float(b)
+        func = lambda b, a: 1 - fancy_division(float(a), float(b))
     elif format == 'times':
         # it would be nice if we could add an "x" after the number, but eh...
         row_num_fmt = '.2f'
-        func = lambda b, a: float(a) / float(b)
+        func = lambda b, a: fancy_division(float(a), float(b))
     else:
         assert False, "shouldn't happen"
 
