@@ -37,7 +37,7 @@ Print the titles of the unread entries::
 
 """
 
-__version__ = '1.13'
+__version__ = '1.18'
 
 from .core import Reader, make_reader
 
@@ -52,6 +52,8 @@ from .types import (
     FeedCounts,
     EntryCounts,
     EntrySearchCounts,
+    UpdateResult,
+    UpdatedFeed,
 )
 
 from .exceptions import (
@@ -64,10 +66,13 @@ from .exceptions import (
     EntryNotFoundError,
     MetadataError,
     MetadataNotFoundError,
+    FeedMetadataNotFoundError,
     StorageError,
     SearchError,
     SearchNotEnabledError,
     InvalidSearchQueryError,
+    PluginError,
+    InvalidPluginError,
 )
 
 
