@@ -32,7 +32,6 @@ from reader import InvalidSearchQueryError
 from reader import ParseError
 from reader import ReaderError
 from reader._plugins import Loader
-from reader._plugins import LoaderError
 
 blueprint = Blueprint('reader', __name__)
 
@@ -48,7 +47,7 @@ got_preview_parse_error = signals.signal('preview-parse-error')
 def get_reader():
     if not hasattr(g, 'reader'):
         g.reader = current_app.config['READER_CONFIG'].make_reader(
-            'app', plugin_loader_cls=FlaskPluginLoader
+            'app', plugin_loader=current_app.plugin_loader
         )
     return g.reader
 
@@ -272,7 +271,7 @@ def preview():
     # TODO: maybe cache stuff
 
     reader = current_app.config['READER_CONFIG'].make_reader(
-        'default', url=':memory:', plugin_loader_cls=FlaskPluginLoader
+        'default', url=':memory:', plugin_loader=current_app.plugin_loader
     )
 
     reader.add_feed(url)
@@ -368,7 +367,7 @@ def metadata():
     if not feed:
         abort(404)
 
-    metadata = reader.iter_feed_metadata(feed_url)
+    metadata = reader.get_feed_metadata(feed_url)
 
     # Ensure flashed messages get removed from the session.
     # https://github.com/lemon24/reader/issues/81
@@ -403,16 +402,19 @@ def tags():
     with_counts = request.args.get('counts')
     with_counts = {None: None, 'no': False, 'yes': True}[with_counts]
 
-    tags = (
-        (
-            tag,
-            reader.get_feed_counts(tags=[tag]) if with_counts else None,
-            reader.get_entry_counts(feed_tags=[tag]) if with_counts else None,
-        )
-        for tag in itertools.chain([True, False], reader.get_feed_tags())
-    )
+    def iter_tags():
+        for tag in itertools.chain([None, True, False], reader.get_feed_tags()):
+            feed_counts = None
+            entry_counts = None
 
-    return render_template('tags.html', tags=tags)
+            if with_counts:
+                tags_arg = [tag] if tag is not None else tag
+                feed_counts = reader.get_feed_counts(tags=tags_arg)
+                entry_counts = reader.get_entry_counts(feed_tags=tags_arg)
+
+            yield tag, feed_counts, entry_counts
+
+    return render_template('tags.html', tags=iter_tags())
 
 
 form_api = APIThing(blueprint, '/form-api', 'form_api')
@@ -436,7 +438,7 @@ def readererror_to_apierror(*args):
 def mark_as_read(data):
     feed_url = data['feed-url']
     entry_id = data['entry-id']
-    get_reader().mark_as_read((feed_url, entry_id))
+    get_reader().mark_entry_as_read((feed_url, entry_id))
 
 
 @form_api
@@ -444,7 +446,7 @@ def mark_as_read(data):
 def mark_as_unread(data):
     feed_url = data['feed-url']
     entry_id = data['entry-id']
-    get_reader().mark_as_unread((feed_url, entry_id))
+    get_reader().mark_entry_as_unread((feed_url, entry_id))
 
 
 @form_api(really=True)
@@ -453,7 +455,7 @@ def mark_all_as_read(data):
     feed_url = data['feed-url']
     entry_id = json.loads(data['entry-id'])
     for entry_id in entry_id:
-        get_reader().mark_as_read((feed_url, entry_id))
+        get_reader().mark_entry_as_read((feed_url, entry_id))
 
 
 @form_api(really=True)
@@ -462,7 +464,7 @@ def mark_all_as_unread(data):
     feed_url = data['feed-url']
     entry_id = json.loads(data['entry-id'])
     for entry_id in entry_id:
-        get_reader().mark_as_unread((feed_url, entry_id))
+        get_reader().mark_entry_as_unread((feed_url, entry_id))
 
 
 @form_api
@@ -470,7 +472,7 @@ def mark_all_as_unread(data):
 def mark_as_important(data):
     feed_url = data['feed-url']
     entry_id = data['entry-id']
-    get_reader().mark_as_important((feed_url, entry_id))
+    get_reader().mark_entry_as_important((feed_url, entry_id))
 
 
 @form_api
@@ -478,14 +480,14 @@ def mark_as_important(data):
 def mark_as_unimportant(data):
     feed_url = data['feed-url']
     entry_id = data['entry-id']
-    get_reader().mark_as_unimportant((feed_url, entry_id))
+    get_reader().mark_entry_as_unimportant((feed_url, entry_id))
 
 
 @form_api(really=True)
 @readererror_to_apierror()
 def delete_feed(data):
     feed_url = data['feed-url']
-    get_reader().remove_feed(feed_url)
+    get_reader().delete_feed(feed_url)
 
 
 @form_api
@@ -510,7 +512,7 @@ def update_feed_title(data):
 def add_metadata(data):
     feed_url = data['feed-url']
     key = data['key']
-    get_reader().set_feed_metadata(feed_url, key, None)
+    get_reader().set_feed_metadata_item(feed_url, key, None)
 
 
 @form_api
@@ -522,7 +524,7 @@ def update_metadata(data):
         value = yaml.safe_load(data['value'])
     except yaml.YAMLError as e:
         raise APIError("invalid JSON: {}".format(e), (feed_url, key))
-    get_reader().set_feed_metadata(feed_url, key, value)
+    get_reader().set_feed_metadata_item(feed_url, key, value)
 
 
 @form_api
@@ -530,7 +532,7 @@ def update_metadata(data):
 def delete_metadata(data):
     feed_url = data['feed-url']
     key = data['key']
-    get_reader().delete_feed_metadata(feed_url, key)
+    get_reader().delete_feed_metadata_item(feed_url, key)
 
 
 @form_api
@@ -572,6 +574,16 @@ def disable_feed_updates(data):
     get_reader().disable_feed_updates(feed_url)
 
 
+@form_api
+@readererror_to_apierror()
+def update_feed(data):
+    # TODO: feed updates should happen in the background
+    # (otherwise we're tying up a worker);
+    # acceptable only because /preview does it as well
+    feed_url = data['feed-url']
+    get_reader().update_feed(feed_url)
+
+
 # for some reason, @blueprint.app_template_global does not work
 @blueprint.app_template_global()
 def additional_enclosure_links(enclosure, entry):
@@ -580,13 +592,6 @@ def additional_enclosure_links(enclosure, entry):
         yield from func(enclosure, entry)
 
 
-class FlaskPluginLoader(Loader):
-    def handle_error(self, exception, cause):
-        current_app.logger.exception(
-            "%s; original traceback follows", exception, exc_info=cause or exception
-        )
-
-
 def create_app(config):
     app = Flask(__name__)
     app.secret_key = 'secret'
@@ -594,18 +599,27 @@ def create_app(config):
     app.config['READER_CONFIG'] = config
     app.teardown_appcontext(close_db)
 
-    # Force reader plugins to load, so we can fail fast for import errors
-    # (although some errors may just be logged while we use FlaskPluginLoader).
-    with app.app_context():
-        get_reader()
-
     app.register_blueprint(blueprint)
 
     # NOTE: this is part of the app extension API
     app.reader_additional_enclosure_links = []
 
-    # app_context() needed for logging to work.
+    app.plugin_loader = loader = Loader()
+
+    def log_exception(message, cause):
+        app.logger.exception("%s; original traceback follows", message, exc_info=cause)
+
+    # Don't raise exceptions for plugins, just log.
+    # Does it make sense to keep going after initializing a plugin fails?
+    # How do we know the target isn't left in a bad state?
+    loader.handle_import_error = log_exception
+    loader.handle_init_error = log_exception
+
+    # Fail fast for reader plugin import/init errors
+    # (although depending on the handler they may just be logged).
     with app.app_context():
-        FlaskPluginLoader(config.merged('app').get('plugins', {})).load_plugins(app)
+        get_reader()
+
+    loader.init(app, config.merged('app').get('plugins', {}))
 
     return app
