@@ -2,17 +2,22 @@ import functools
 import json
 import logging
 import os.path
+import sys
 import traceback
+from datetime import datetime
 
 import click
 import yaml
 
 import reader
+from . import ParseError
+from . import ReaderError
 from . import StorageError
 from ._config import make_reader_config
 from ._config import make_reader_from_config
 from ._plugins import LoaderError
 from ._sqlite_utils import DebugConnection
+from ._utils import nullcontext
 
 
 APP_NAME = reader.__name__
@@ -61,7 +66,7 @@ def make_reader_with_plugins(*, debug_storage=False, **kwargs):
     try:
         return make_reader_from_config(**kwargs)
     except StorageError as e:
-        abort("{}: {}: {}", kwargs['url'], e, e.__cause__)
+        abort("{}: {}", kwargs['url'], e)
     except LoaderError as e:
         abort("{}; original traceback follows\n\n{}", e, format_tb(e.__cause__ or e))
     except Exception as e:
@@ -69,6 +74,8 @@ def make_reader_with_plugins(*, debug_storage=False, **kwargs):
 
 
 def setup_logging(verbose):
+    if verbose < 0:
+        return
     if verbose == 0:
         level = logging.WARNING
     elif verbose == 1:
@@ -84,18 +91,25 @@ def setup_logging(verbose):
     logging.getLogger('reader').addHandler(handler)
 
 
-def log_verbose(fn):
-    @click.option('-v', '--verbose', count=True)
-    @functools.wraps(fn)
-    def wrapper(*args, verbose, **kwargs):
-        setup_logging(verbose)
-        return fn(*args, **kwargs)
+def make_log_verbose(expose_value=False, initial=0):
+    def log_verbose(fn):
+        @click.option('-v', '--verbose', count=True)
+        @functools.wraps(fn)
+        def wrapper(*args, **kwargs):
+            setup_logging(kwargs['verbose'] + initial)
+            if not expose_value:
+                del kwargs['verbose']
+            return fn(*args, **kwargs)
 
-    return wrapper
+        return wrapper
+
+    return log_verbose
+
+
+log_verbose = make_log_verbose()
 
 
 def log_command(fn):
-    @log_verbose
     @functools.wraps(fn)
     def wrapper(*args, **kwargs):
         ctx = click.get_current_context()
@@ -112,13 +126,14 @@ def log_command(fn):
             rv = fn(*args, **kwargs)
             log.info("command finished successfully")
             return rv
-
         except Exception as e:
             log.critical(
                 "command failed due to unexpected error: %s; traceback follows",
                 e,
                 exc_info=True,
             )
+            if not isinstance(e, ReaderError):
+                raise
             click.get_current_context().exit(1)
 
     return wrapper
@@ -182,7 +197,7 @@ def pass_reader(fn):
 @config_option(
     '--config',
     envvar=reader._CONFIG_ENVVAR,
-    help=f"Path to the reader config.",
+    help="Path to the reader config.",
     default=get_default_config_path(),
     show_default=True,
 )
@@ -235,7 +250,46 @@ def add(reader, url, update):
 @pass_reader
 def remove(reader, url):
     """Remove an existing feed."""
-    reader.remove_feed(url)
+    reader.delete_feed(url)
+
+
+def red(text):
+    return click.style(str(text), fg='bright_red')
+
+
+def green(text):
+    return click.style(str(text), fg='bright_green')
+
+
+def get_update_status(value):
+    if value is None:
+        return None
+    if isinstance(value, Exception):
+        return False
+    if not (value.new or value.updated):
+        return None
+    return True
+
+
+def iter_update_status(it, length):
+    start = datetime.now()
+
+    for i, (url, value) in enumerate(it):
+        elapsed = datetime.now() - start
+        pos = f"{i}/{length or '?'}"
+
+        update_status = get_update_status(value)
+
+        if update_status is None:
+            status = 'not modified'
+        elif not update_status:
+            status = red(value)
+        else:
+            status = green(f"{value.new} new, {value.updated} updated")
+
+        click.echo(f"{elapsed}\t{pos}\t{url}\t{status}")
+
+        yield url, value
 
 
 @cli.command()
@@ -250,18 +304,94 @@ def remove(reader, url):
     show_default=True,
     help="Number of threads to use when getting the feeds.",
 )
+@make_log_verbose(True, -2)
 @log_command
 @pass_reader
-def update(reader, url, new_only, workers):
+def update(reader, url, new_only, workers, verbose):
     """Update one or all feeds.
 
     If URL is not given, update all the feeds.
 
+    Verbosity works like this:
+
+    \b
+        : progress bar + final status
+        -v: + lines
+        -vv: + warnings
+        -vvv: + info
+        -vvvv: + debug
+
     """
     if url:
-        reader.update_feed(url)
+
+        def make_it():
+            try:
+                yield url, reader.update_feed(url)
+            except ParseError as e:
+                yield url, e
+
+        it = make_it()
     else:
-        reader.update_feeds(new_only=new_only, workers=workers)
+        it = reader.update_feeds_iter(new_only=new_only, workers=workers)
+
+    ok_count = 0
+    not_modified_count = 0
+    error_count = 0
+    new_count = 0
+    updated_count = 0
+
+    def feed_stats(width=None):
+        if not width:
+            width, _ = click.get_terminal_size()
+        if width < 80:
+            return ''
+        if width < 105:
+            return f"{green(ok_count)}/{red(error_count)}/{not_modified_count}"
+        return (
+            f"{green(f'{ok_count} ok') if ok_count else '0 ok'}, "
+            f"{red(f'{error_count} error') if error_count else '0 error'}, "
+            f"{not_modified_count} not modified"
+        )
+
+    if url:
+        length = 1
+    else:
+        if not new_only:
+            length = reader.get_feed_counts(updates_enabled=True).total
+        else:
+            # TODO: pending https://github.com/lemon24/reader/issues/217
+            length = None
+
+    if not verbose:
+        bar_context = click.progressbar(
+            it,
+            length=length,
+            label='update',
+            show_pos=True,
+            show_eta=True,
+            item_show_func=lambda _: feed_stats(),
+            file=sys.stderr,
+        )
+
+    else:
+        bar_context = nullcontext(iter_update_status(it, length))
+
+    try:
+        with bar_context as bar:
+            for _, value in bar:
+                update_status = get_update_status(value)
+                if update_status is None:
+                    not_modified_count += 1
+                elif not update_status:
+                    error_count += 1
+                else:
+                    ok_count += 1
+                    new_count += value.new
+                    updated_count += value.updated
+    finally:
+        click.echo(
+            f"{feed_stats(9999)}; entries: {new_count} new, {updated_count} updated"
+        )
 
 
 @cli.group('list')
@@ -318,6 +448,7 @@ def search_disable(reader):
 
 
 @search.command('update')
+@log_verbose
 @log_command
 @pass_reader
 def search_update(reader):
