@@ -36,6 +36,8 @@ from ._http_utils import parse_options_header
 from ._http_utils import unparse_accept_header
 from ._requests_utils import SessionHooks
 from ._requests_utils import SessionWrapper
+from ._requests_utils import TimeoutHTTPAdapter
+from ._requests_utils import TimeoutType
 from ._types import EntryData
 from ._types import FeedData
 from ._types import ParsedFeed
@@ -43,7 +45,6 @@ from ._url_utils import extract_path
 from ._url_utils import is_windows_device_file
 from ._url_utils import normalize_url
 from ._url_utils import resolve_root
-from .exceptions import _NotModified
 from .exceptions import ParseError
 from .types import Content
 from .types import Enclosure
@@ -70,7 +71,7 @@ class RetrieverType(Protocol):
         http_etag: Optional[str],
         http_last_modified: Optional[str],
         http_accept: Optional[str],
-    ) -> ContextManager[RetrieveResult]:  # pragma: no cover
+    ) -> ContextManager[Optional[RetrieveResult]]:  # pragma: no cover
         ...
 
 
@@ -91,8 +92,14 @@ class AwareParserType(ParserType, Protocol):
         ...
 
 
-def default_parser(feed_root: Optional[str] = None) -> 'Parser':
+SESSION_TIMEOUT = (3.05, 60)
+
+
+def default_parser(
+    feed_root: Optional[str] = None, session_timeout: TimeoutType = SESSION_TIMEOUT
+) -> 'Parser':
     parser = Parser()
+    parser.session_timeout = session_timeout
 
     http_retriever = HTTPRetriever(parser.make_session)
     parser.mount_retriever('https://', http_retriever)
@@ -124,13 +131,14 @@ class Parser:
         self.parsers_by_mime_type: Dict[str, List[Tuple[float, ParserType]]] = {}
         self.parsers_by_url: Dict[str, ParserType] = {}
         self.session_hooks = SessionHooks()
+        self.session_timeout: TimeoutType = None
 
     def __call__(
         self,
         url: str,
         http_etag: Optional[str] = None,
         http_last_modified: Optional[str] = None,
-    ) -> ParsedFeed:
+    ) -> Optional[ParsedFeed]:
         parser = self.get_parser_by_url(url)
 
         http_accept: Optional[str]
@@ -148,6 +156,9 @@ class Parser:
         retriever = self.get_retriever(url)
 
         with retriever(url, http_etag, http_last_modified, http_accept) as result:
+            if not result:
+                return None
+
             if not parser:
                 mime_type = result.mime_type
                 if not mime_type:
@@ -227,8 +238,13 @@ class Parser:
 
     def make_session(self) -> SessionWrapper:
         session = SessionWrapper(hooks=self.session_hooks.copy())
+
+        session.session.mount('https://', TimeoutHTTPAdapter(self.session_timeout))
+        session.session.mount('http://', TimeoutHTTPAdapter(self.session_timeout))
+
         if self.user_agent:
             session.session.headers['User-Agent'] = self.user_agent
+
         return session
 
 
@@ -236,7 +252,7 @@ class Parser:
 def wrap_exceptions(url: str, when: str) -> Iterator[None]:
     try:
         yield
-    except (ParseError, _NotModified):
+    except ParseError:
         # reader exceptions are pass-through
         raise
     except OSError as e:
@@ -322,13 +338,12 @@ class HTTPRetriever:
         http_etag: Optional[str] = None,
         http_last_modified: Optional[str] = None,
         http_accept: Optional[str] = None,
-    ) -> Iterator[RetrieveResult]:
+    ) -> Iterator[Optional[RetrieveResult]]:
         request_headers = {}
         if http_accept:
             request_headers['Accept'] = http_accept
 
         # TODO: maybe share the session in the parser?
-        # TODO: timeouts!
 
         with self.make_session() as session:
             with wrap_exceptions(url, "while getting feed"):
@@ -341,6 +356,11 @@ class HTTPRetriever:
                     stream=True,
                 )
 
+            if response.status_code == 304:
+                response.close()
+                yield None
+                return
+
             response_headers = response.headers.copy()
             response_headers.setdefault('content-location', response.url)
 
@@ -387,9 +407,6 @@ def _caching_get(
     except Exception as e:
         raise ParseError(url, message="bad HTTP status code") from e
 
-    if response.status_code == 304:
-        raise _NotModified(url)
-
     http_etag = response.headers.get('ETag', http_etag)
     http_last_modified = response.headers.get('Last-Modified', http_last_modified)
 
