@@ -19,7 +19,6 @@ from typing import no_type_check
 from typing import Optional
 from typing import Sequence
 from typing import Tuple
-from typing import TYPE_CHECKING
 from typing import TypeVar
 
 from typing_extensions import Protocol
@@ -192,7 +191,11 @@ class RequirementError(DBError):
     display_name = "database requirement error"
 
 
-db_errors = [DBError, SchemaVersionError, RequirementError]
+class IdError(DBError):
+    display_name = "application id error"
+
+
+db_errors = [DBError, SchemaVersionError, IntegrityError, RequirementError]
 
 
 class _DBFunction(Protocol):  # pragma: no cover
@@ -201,37 +204,43 @@ class _DBFunction(Protocol):  # pragma: no cover
 
 
 @dataclass
-class HeavyMigration:
+class NewMigration:
 
     # mypy will complain if we use Callable[[sqlite3.Connection], None].
     # TODO: get rid of _DBFunction when https://github.com/python/mypy/issues/5485 is resolved?
     create: _DBFunction
-    version: int
+    version: int  # must be positive
     migrations: Dict[int, _DBFunction]
-
-    @staticmethod
-    def get_version(db: sqlite3.Connection) -> Optional[int]:
-        try:
-            # TODO: this assignment should fail with DBError
-            (version,) = db.execute("SELECT MAX(version) FROM version;").fetchone()
-        except sqlite3.OperationalError as e:
-            if "no such table" in str(e).lower():
-                return None
-            raise SchemaVersionError(f"cannot get current version: {e}") from e
-        assert isinstance(version, int)
-        return version
+    id: int = 0
 
     def migrate(self, db: sqlite3.Connection) -> None:
+        # pseudo-code for how the application_id is handled:
+        # https://github.com/lemon24/reader/issues/211#issuecomment-778392468
+        # unlike there, we allow bypassing it for testing
+
         with foreign_keys_off(db), ddl_transaction(db):
+            if self.id:
+                id = self.get_id(db)
+                if id and id != self.id:
+                    raise IdError(f"invalid id: 0x{id:x}")
+
             version = self.get_version(db)
 
-            if version is None:
+            if not version:
+                # avoid clobbering a database with application_id
+                if table_count(db) != 0:
+                    # TODO: maybe use a custom exception here?
+                    raise DBError("database with no version already has tables")
+
                 self.create(db)
-                db.execute("CREATE TABLE version (version INTEGER NOT NULL);")
-                db.execute("INSERT INTO version VALUES (?);", (self.version,))
+                self.set_version(db, self.version)
+                self.set_id(db, self.id)
                 return
 
             if version == self.version:
+                if self.id:
+                    if not id:
+                        raise IdError("database with version has missing id")
                 return
 
             if version > self.version:
@@ -239,6 +248,15 @@ class HeavyMigration:
 
             # version < self.version
 
+            # the actual migration code;
+            #
+            # might clobber a database if all of the below are true:
+            #
+            # * an application_id was not used from the start
+            # * the database has a non-zero version which predates
+            #   the migration which set application_id
+            # * all of the migrations succeed
+
             for from_version in range(version, self.version):
                 to_version = from_version + 1
                 migration = self.migrations.get(from_version)
@@ -249,10 +267,7 @@ class HeavyMigration:
                         f"later than {version}"
                     )
 
-                db.execute(
-                    "UPDATE version SET version = :to_version;",
-                    dict(to_version=to_version),
-                )
+                self.set_version(db, to_version)
                 migration(db)
 
                 try:
@@ -262,6 +277,95 @@ class HeavyMigration:
                         f"after migrating to version {to_version}: {e}"
                     ) from None
 
+            if self.id:
+                id = self.get_id(db)
+                if id != self.id:
+                    raise IdError(f"missing or invalid id after migration: 0x{id:x}")
+
+    @staticmethod
+    def get_version(db: sqlite3.Connection) -> int:
+        return get_int_pragma(db, 'user_version')
+
+    @staticmethod
+    def set_version(db: sqlite3.Connection, version: int) -> None:
+        set_int_pragma(db, 'user_version', version)
+
+    @staticmethod
+    def get_id(db: sqlite3.Connection) -> int:
+        return get_int_pragma(db, 'application_id')
+
+    @staticmethod
+    def set_id(db: sqlite3.Connection, id: int) -> None:
+        set_int_pragma(db, 'application_id', id)
+
+
+def get_int_pragma(db: sqlite3.Connection, pragma: str) -> int:
+    (value,) = db.execute(f"PRAGMA {pragma};").fetchone()
+    assert isinstance(value, int), value  # for mypy
+    return value
+
+
+def set_int_pragma(
+    db: sqlite3.Connection, pragma: str, value: int, lower_bound: int = 0
+) -> None:
+    if not isinstance(value, int):
+        raise ValueError(f"{pragma} must be an integer, got {value!r}")
+    if lower_bound is not None and value < lower_bound:
+        raise ValueError(f"{pragma} must be >={lower_bound}, got {value!r}")
+
+    db.execute(f"PRAGMA {pragma} = {value};")
+
+
+def table_count(db: sqlite3.Connection) -> int:
+    (value,) = db.execute("select count(*) from sqlite_master;").fetchone()
+    assert isinstance(value, int), value  # for mypy
+    return value
+
+
+class OldMigration(NewMigration):
+
+    # TODO: delete in 2.0
+
+    @staticmethod
+    def get_version(db: sqlite3.Connection) -> int:
+        try:
+            # TODO: this assignment should fail with DBError
+            (version,) = db.execute("SELECT MAX(version) FROM version;").fetchone()
+        except sqlite3.OperationalError as e:
+            if "no such table" in str(e).lower():
+                return 0
+            raise SchemaVersionError(f"cannot get current version: {e}") from e
+        assert isinstance(version, int), version
+        return version
+
+    @classmethod
+    def set_version(
+        cls, db: sqlite3.Connection, version: int
+    ) -> None:  # pragma: no cover
+        raise NotImplementedError
+
+    @staticmethod
+    def del_version(db: sqlite3.Connection) -> None:
+        db.execute("DROP TABLE IF EXISTS version;")
+
+
+class HeavyMigration(NewMigration):
+
+    """Transition wrapper to assist with the switch from from storing
+    the schema version in a "version" table to using "PRAGMA user_version".
+
+    TODO: delete in 2.0, and rename NewMigration to HeavyMigration
+
+    """
+
+    def migrate(self, db: sqlite3.Connection) -> None:
+        old_version = OldMigration.get_version(db)
+        if old_version:
+            with ddl_transaction(db):
+                OldMigration.del_version(db)
+                super().set_version(db, old_version)
+        super().migrate(db)
+
 
 def require_version(db: sqlite3.Connection, version_info: Tuple[int, ...]) -> None:
     with closing(db.cursor()) as cursor:
@@ -295,6 +399,7 @@ def setup_db(
     create: _DBFunction,
     version: int,
     migrations: Dict[int, _DBFunction],
+    id: int,
     minimum_sqlite_version: Tuple[int, ...],
     required_sqlite_compile_options: Sequence[str] = (),
     wal_enabled: Optional[bool] = None,
@@ -319,7 +424,7 @@ def setup_db(
             else:
                 cursor.execute("PRAGMA journal_mode = DELETE;")
 
-    migration = HeavyMigration(create, version, migrations)
+    migration = HeavyMigration(create, version, migrations, id)
     migration.migrate(db)
 
 
@@ -331,49 +436,6 @@ def rowcount_exactly_one(
     assert cursor.rowcount == 1, "shouldn't have more than 1 row"
 
 
-# TODO: maybe move this to _sql_utils,
-# since DB can be anything with .execute(), but a Query is always required;
-# leaving here for now because _sql_utils.py has type checking disabled
-
-_T = TypeVar('_T')
-_U = TypeVar('_U')
-
-if TYPE_CHECKING:  # pragma: no cover
-    from ._sql_utils import Query
-
-
-def paginated_query(
-    db: sqlite3.Connection,
-    query: 'Query',
-    # FIXME: Any should be SQLiteType
-    context: Dict[str, Any],
-    value_factory: Callable[[Tuple[Any, ...]], _T],
-    chunk_size: Optional[int],
-    last: _U,
-) -> Iterator[Tuple[_T, _U]]:
-    # We need the cast to/from _U until we find a better way of typing last.
-
-    if chunk_size:
-        query.LIMIT(":chunk_size")
-        context['chunk_size'] = chunk_size
-    if last:
-        query.add_last()
-        context.update(query.last_params(cast(Optional[Tuple[Any, ...]], last)))
-
-    rv = (
-        (value_factory(t), cast(_U, query.extract_last(t)))
-        for t in db.execute(str(query), context)
-    )
-
-    # Equivalent to using @returns_iter_list, except when we don't have
-    # a chunk_size (which disables pagination, but can block the database).
-    # TODO: If we don't expose chunk_size, why have this special case?
-    if chunk_size:
-        return iter(list(rv))
-
-    return rv
-
-
 # BEGIN DebugConnection
 
 # No type annotations or coverage for this;
