@@ -10,15 +10,17 @@ from typing import Any
 from typing import Dict
 from typing import Iterable
 from typing import Mapping
+from typing import NamedTuple
 from typing import Optional
 from typing import Sequence
 from typing import Tuple
 from typing import Type
 from typing import TypeVar
 
+from ._sql_utils import BaseQuery
+from ._sql_utils import paginated_query
 from ._sql_utils import Query
 from ._sqlite_utils import DBError
-from ._sqlite_utils import paginated_query
 from ._sqlite_utils import rowcount_exactly_one
 from ._sqlite_utils import setup_db as setup_sqlite_db
 from ._sqlite_utils import wrap_exceptions
@@ -34,10 +36,13 @@ from ._utils import chunks
 from ._utils import exactly_one
 from ._utils import join_paginated_iter
 from ._utils import zero_or_one
+from .exceptions import EntryMetadataNotFoundError
 from .exceptions import EntryNotFoundError
 from .exceptions import FeedExistsError
+from .exceptions import FeedMetadataNotFoundError
 from .exceptions import FeedNotFoundError
 from .exceptions import MetadataNotFoundError
+from .exceptions import ReaderError
 from .exceptions import StorageError
 from .types import Content
 from .types import Enclosure
@@ -50,6 +55,8 @@ from .types import FeedCounts
 from .types import FeedSortOrder
 from .types import JSONType
 
+APPLICATION_ID = int(''.join(f'{ord(c):x}' for c in 'read'), 16)
+
 
 log = logging.getLogger('reader')
 
@@ -79,6 +86,7 @@ def create_feeds(db: sqlite3.Connection, name: str = 'feeds') -> None:
             user_title TEXT,  -- except this one, which comes from reader
             http_etag TEXT,
             http_last_modified TEXT,
+            data_hash BLOB,  -- derived from feed data
 
             -- reader data
             stale INTEGER NOT NULL DEFAULT 0,
@@ -88,7 +96,7 @@ def create_feeds(db: sqlite3.Connection, name: str = 'feeds') -> None:
             last_exception TEXT
 
             -- NOTE: when adding new fields, check if they should be set
-            -- to their defalt value in change_feed_url()
+            -- to their default value in change_feed_url()
 
         );
         """
@@ -112,6 +120,8 @@ def create_entries(db: sqlite3.Connection, name: str = 'entries') -> None:
             content TEXT,
             enclosures TEXT,
             original_feed TEXT,  -- null if the feed was never moved
+            data_hash BLOB,  -- derived from entry data
+            data_hash_changed INTEGER,  -- metadata about data_hash
 
             -- reader data
             read INTEGER NOT NULL DEFAULT 0,
@@ -162,6 +172,21 @@ def create_feed_tags(db: sqlite3.Connection) -> None:
     )
 
 
+class SchemaInfo(NamedTuple):
+    table_prefix: str
+    id_columns: Tuple[str, ...]
+    not_found_exc: Type[ReaderError]
+    metadata_not_found_exc: Type[MetadataNotFoundError]
+
+
+SCHEMA_INFO = {
+    1: SchemaInfo('feed_', ('feed',), FeedNotFoundError, FeedMetadataNotFoundError),
+    2: SchemaInfo(
+        'entry_', ('feed', 'id'), EntryNotFoundError, EntryMetadataNotFoundError
+    ),
+}
+
+
 def create_indexes(db: sqlite3.Connection) -> None:
     # Speed up get_entries() queries that use apply_recent().
     db.execute(
@@ -259,11 +284,33 @@ def update_from_25_to_26(db: sqlite3.Connection) -> None:  # pragma: no cover
     )
 
 
+def update_from_26_to_27(db: sqlite3.Connection) -> None:  # pragma: no cover
+    # for https://github.com/lemon24/reader/issues/211
+    db.execute(f"PRAGMA application_id = {APPLICATION_ID};")
+
+
+def update_from_27_to_28(db: sqlite3.Connection) -> None:  # pragma: no cover
+    # for https://github.com/lemon24/reader/issues/179
+
+    db.execute("ALTER TABLE feeds ADD COLUMN data_hash BLOB;")
+    db.execute("ALTER TABLE entries ADD COLUMN data_hash BLOB;")
+
+    # force the hash to be set on the next full update;
+    # otherwise, for some feeds it'll be set only after
+    # their caching headers change, which is less predictable
+    db.execute("UPDATE feeds SET stale = 1;")
+
+
+def update_from_28_to_29(db: sqlite3.Connection) -> None:  # pragma: no cover
+    # for https://github.com/lemon24/reader/issues/225
+    db.execute("ALTER TABLE entries ADD COLUMN data_hash_changed INTEGER;")
+
+
 def setup_db(db: sqlite3.Connection, wal_enabled: Optional[bool]) -> None:
     return setup_sqlite_db(
         db,
         create=create_db,
-        version=26,
+        version=29,
         migrations={
             # 1-9 removed before 0.1 (last in e4769d8ba77c61ec1fe2fbe99839e1826c17ace7)
             # 10-16 removed before 1.0 (last in 618f158ebc0034eefb724a55a84937d21c93c1a7)
@@ -280,7 +327,11 @@ def setup_db(db: sqlite3.Connection, wal_enabled: Optional[bool]) -> None:
             # for https://github.com/lemon24/reader/issues/134#issuecomment-722454963
             24: create_indexes,
             25: update_from_25_to_26,
+            26: update_from_26_to_27,
+            27: update_from_27_to_28,
+            28: update_from_28_to_29,
         },
+        id=APPLICATION_ID,
         # Row value support was added in 3.15.
         # TODO: Remove the Search.update() check once this gets bumped to >=3.18.
         minimum_sqlite_version=(3, 15),
@@ -384,7 +435,7 @@ class Storage:
                 raise FeedExistsError(url)
 
     @wrap_exceptions(StorageError)
-    def remove_feed(self, url: str) -> None:
+    def delete_feed(self, url: str) -> None:
         with self.db:
             cursor = self.db.execute(
                 "DELETE FROM feeds WHERE url = :url;", dict(url=url)
@@ -477,37 +528,9 @@ class Storage:
         chunk_size: Optional[int] = None,
         last: Optional[_T] = None,
     ) -> Iterable[Tuple[Feed, Optional[_T]]]:
-        query = (
-            Query()
-            .SELECT(
-                'url',
-                'updated',
-                'title',
-                'link',
-                'author',
-                'user_title',
-                'added',
-                'last_updated',
-                'last_exception',
-                'updates_enabled',
-            )
-            .FROM("feeds")
-        )
-
-        context = apply_feed_filter_options(query, filter_options)
-
-        # sort by url at the end to make sure the order is deterministic
-        if sort == 'title':
-            query.SELECT(("kinda_title", "lower(coalesce(user_title, title))"))
-            query.scrolling_window_order_by("kinda_title", "url")
-        elif sort == 'added':
-            query.SELECT("added")
-            query.scrolling_window_order_by("added", "url", desc=True)
-        else:
-            assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
-
+        query, context = make_get_feeds_query(filter_options, sort)
         yield from paginated_query(
-            self.db, query, context, feed_factory, chunk_size, last
+            self.db, query, context, chunk_size, last, feed_factory
         )
 
     @wrap_exceptions(StorageError)
@@ -554,6 +577,7 @@ class Storage:
                     'stale',
                     'last_updated',
                     ('last_exception', 'last_exception IS NOT NULL'),
+                    'data_hash',
                 )
                 .FROM("feeds")
             )
@@ -572,7 +596,7 @@ class Storage:
             query.scrolling_window_order_by("url")
 
             yield from paginated_query(
-                self.db, query, context, FeedForUpdate._make, chunk_size, last
+                self.db, query, context, chunk_size, last, FeedForUpdate._make
             )
 
         yield from join_paginated_iter(inner, self.chunk_size)
@@ -587,14 +611,14 @@ class Storage:
                 context = dict(feed_url=feed_url, id=id)
                 row = self.db.execute(
                     """
-                    SELECT updated
+                    SELECT updated, data_hash, data_hash_changed
                     FROM entries
                     WHERE feed = :feed_url
                         AND id = :id;
                     """,
                     context,
                 ).fetchone()
-                yield EntryForUpdate(row[0]) if row else None
+                yield EntryForUpdate._make(row) if row else None
 
     def _get_entries_for_update_one_query(
         self, entries: Sequence[Tuple[str, str]]
@@ -613,7 +637,9 @@ class Storage:
                 )
             SELECT
                 entries.id IS NOT NULL,
-                entries.updated
+                entries.updated,
+                entries.data_hash,
+                entries.data_hash_changed
             FROM input
             LEFT JOIN entries
                 ON (input.id, input.feed) == (entries.id, entries.feed);
@@ -623,7 +649,9 @@ class Storage:
 
         # This can't be a generator because we need to get OperationalError
         # in this function (so get_entries_for_update() below can catch it).
-        return (EntryForUpdate(updated) if exists else None for exists, updated in rows)
+        return (
+            EntryForUpdate._make(rest) if exists else None for exists, *rest in rows
+        )
 
     @wrap_exceptions_iter(StorageError)
     def get_entries_for_update(
@@ -736,10 +764,7 @@ class Storage:
         assert feed is not None
         context.pop('last_exception')
 
-        # TODO: can't use context.update(feed._asdict()) because for some tests intent.feed is Feed instead of FeedData
-        context.update(
-            updated=feed.updated, title=feed.title, link=feed.link, author=feed.author
-        )
+        context.update(feed._asdict(), data_hash=feed.hash)
 
         with self.db:
             cursor = self.db.execute(
@@ -752,6 +777,7 @@ class Storage:
                     author = :author,
                     http_etag = :http_etag,
                     http_last_modified = :http_last_modified,
+                    data_hash = :data_hash,
                     stale = 0,
                     last_updated = :last_updated,
                     last_exception = NULL
@@ -796,8 +822,8 @@ class Storage:
     ) -> Mapping[str, Any]:
         context = intent._asdict()
         entry = context.pop('entry')
-        context.update(entry._asdict())
         context.update(
+            entry._asdict(),
             content=(
                 json.dumps([t._asdict() for t in entry.content])
                 if entry.content
@@ -808,6 +834,8 @@ class Storage:
                 if entry.enclosures
                 else None
             ),
+            data_hash=entry.hash,
+            data_hash_changed=context.pop('hash_changed'),
         )
         return context
 
@@ -861,7 +889,9 @@ class Storage:
                         last_updated,
                         first_updated_epoch,
                         feed_order,
-                        original_feed
+                        original_feed,
+                        data_hash,
+                        data_hash_changed
                     ) VALUES (
                         :id,
                         :feed_url,
@@ -890,7 +920,9 @@ class Storage:
                             WHERE id = :id AND feed = :feed_url
                         )),
                         :feed_order,
-                        NULL
+                        NULL, -- original_feed
+                        :data_hash,
+                        :data_hash_changed
                     );
                     """,
                     make_params(),
@@ -985,31 +1017,10 @@ class Storage:
         chunk_size: Optional[int] = None,
         last: Optional[_T] = None,
     ) -> Iterable[Tuple[Entry, Optional[_T]]]:
-        # See this issue for some thoughts on the sort='random' implementation:
-        # https://github.com/lemon24/reader/issues/105
-
-        query, query_context = make_get_entries_query(filter_options, sort)
-
-        context = dict(
-            recent_threshold=now - self.recent_threshold,
-            **query_context,
-        )
-
-        def value_factory(t: Tuple[Any, ...]) -> Entry:
-            feed = feed_factory(t[0:10])
-            entry = t[10:17] + (
-                tuple(Content(**d) for d in json.loads(t[17])) if t[17] else (),
-                tuple(Enclosure(**d) for d in json.loads(t[18])) if t[18] else (),
-                t[19] == 1,
-                t[20] == 1,
-                t[21],
-                t[22] or feed.url,
-                feed,
-            )
-            return Entry._make(entry)
-
+        query, context = make_get_entries_query(filter_options, sort)
+        context.update(recent_threshold=now - self.recent_threshold)
         yield from paginated_query(
-            self.db, query, context, value_factory, chunk_size, last
+            self.db, query, context, chunk_size, last, entry_factory
         )
 
     @wrap_exceptions(StorageError)
@@ -1042,133 +1053,211 @@ class Storage:
 
         return EntryCounts(*row)
 
-    def iter_feed_metadata(
+    def iter_metadata(
         self,
-        feed_url: str,
+        object_id: Tuple[str, ...],
         key: Optional[str] = None,
     ) -> Iterable[Tuple[str, JSONType]]:
         yield from join_paginated_iter(
-            partial(self.iter_feed_metadata_page, feed_url, key),
+            partial(self.iter_metadata_page, object_id, key),
             self.chunk_size,
         )
 
     @wrap_exceptions_iter(StorageError)
-    def iter_feed_metadata_page(
+    def iter_metadata_page(
         self,
-        feed_url: str,
+        object_id: Tuple[str, ...],
         key: Optional[str] = None,
         chunk_size: Optional[int] = None,
         last: Optional[_T] = None,
     ) -> Iterable[Tuple[Tuple[str, JSONType], Optional[_T]]]:
-        query = (
-            Query()
-            .SELECT("key", "value")
-            .FROM("feed_metadata")
-            .WHERE("feed = :feed_url")
-        )
-        context = dict(feed_url=feed_url)
+        info = SCHEMA_INFO[len(object_id)]
+
+        query = Query().SELECT("key", "value").FROM(f"{info.table_prefix}metadata")
+
+        for column in info.id_columns:
+            query.WHERE(f"{column} = :{column}")
+        context = dict(zip(info.id_columns, object_id))
+
         if key is not None:
             query.WHERE("key = :key")
             context.update(key=key)
 
         query.scrolling_window_order_by("key")
 
-        def value_factory(t: Tuple[Any, ...]) -> Tuple[str, JSONType]:
+        def row_factory(t: Tuple[Any, ...]) -> Tuple[str, JSONType]:
             key, value, *_ = t
             return key, json.loads(value)
 
         yield from paginated_query(
-            self.db, query, context, value_factory, chunk_size, last
+            self.db, query, context, chunk_size, last, row_factory
         )
 
     @wrap_exceptions(StorageError)
-    def set_feed_metadata(self, feed_url: str, key: str, value: JSONType) -> None:
-        with self.db:
-            try:
-                self.db.execute(
-                    """
-                    INSERT OR REPLACE INTO feed_metadata (
-                        feed,
-                        key,
-                        value
-                    ) VALUES (
-                        :feed_url,
-                        :key,
-                        :value_json
-                    );
-                    """,
-                    dict(feed_url=feed_url, key=key, value_json=json.dumps(value)),
-                )
-            except sqlite3.IntegrityError as e:
-                if (
-                    "foreign key constraint failed" not in str(e).lower()
-                ):  # pragma: no cover
-                    raise
-                raise FeedNotFoundError(feed_url)
+    def set_metadata(
+        self, object_id: Tuple[str, ...], key: str, value: JSONType
+    ) -> None:
+        info = SCHEMA_INFO[len(object_id)]
 
-    @wrap_exceptions(StorageError)
-    def delete_feed_metadata(self, feed_url: str, key: str) -> None:
-        with self.db:
-            cursor = self.db.execute(
-                """
-                DELETE FROM feed_metadata
-                WHERE feed = :feed_url AND key = :key;
-                """,
-                dict(feed_url=feed_url, key=key),
+        columns = info.id_columns + ('key', 'value')
+        query = f"""
+            INSERT OR REPLACE INTO {info.table_prefix}metadata (
+                {', '.join(columns)}
+            ) VALUES (
+                {', '.join(('?' for _ in columns))}
             )
-        rowcount_exactly_one(cursor, lambda: MetadataNotFoundError(feed_url, key))
+        """
+        params = object_id + (key, json.dumps(value))
 
-    @wrap_exceptions(StorageError)
-    def add_feed_tag(self, feed_url: str, tag: str) -> None:
         with self.db:
             try:
-                self.db.execute("INSERT INTO feed_tags VALUES (?, ?);", (feed_url, tag))
+                self.db.execute(query, params)
+            except sqlite3.IntegrityError as e:
+                foreign_key_error = "foreign key constraint failed" in str(e).lower()
+                if not foreign_key_error:  # pragma: no cover
+                    raise
+                raise info.not_found_exc(*object_id)
+
+    @wrap_exceptions(StorageError)
+    def delete_metadata(self, object_id: Tuple[str, ...], key: str) -> None:
+        info = SCHEMA_INFO[len(object_id)]
+
+        columns = info.id_columns + ('key',)
+        query = f"""
+            DELETE FROM {info.table_prefix}metadata
+            WHERE (
+                {', '.join(columns)}
+            ) = (
+                {', '.join(('?' for _ in columns))}
+            )
+        """
+        params = object_id + (key,)
+
+        with self.db:
+            cursor = self.db.execute(query, params)
+        rowcount_exactly_one(
+            cursor, lambda: info.metadata_not_found_exc(*object_id, key=key)
+        )
+
+    @wrap_exceptions(StorageError)
+    def add_tag(self, object_id: Tuple[str, ...], tag: str) -> None:
+        info = SCHEMA_INFO[len(object_id)]
+
+        columns = info.id_columns + ('tag',)
+        query = f"""
+            INSERT INTO {info.table_prefix}tags (
+                {', '.join(columns)}
+            ) VALUES (
+                {', '.join(('?' for _ in columns))}
+            )
+        """
+        params = object_id + (tag,)
+
+        with self.db:
+            try:
+                self.db.execute(query, params)
             except sqlite3.IntegrityError as e:
                 if "foreign key constraint failed" in str(e).lower():
-                    raise FeedNotFoundError(feed_url)
+                    raise info.not_found_exc(*object_id)
                 # tag exists is a no-op; it looks like:
                 # "UNIQUE constraint failed: feed_tags.feed, feed_tags.tag"
 
     @wrap_exceptions(StorageError)
-    def remove_feed_tag(self, feed_url: str, tag: str) -> None:
-        with self.db:
-            self.db.execute(
-                "DELETE FROM feed_tags WHERE feed = ? AND tag = ?;",
-                (feed_url, tag),
-            )
+    def remove_tag(self, object_id: Tuple[str, ...], tag: str) -> None:
+        info = SCHEMA_INFO[len(object_id)]
 
-    def get_feed_tags(self, feed_url: Optional[str]) -> Iterable[str]:
+        columns = info.id_columns + ('tag',)
+        query = f"""
+            DELETE FROM {info.table_prefix}tags
+            WHERE (
+                {', '.join(columns)}
+            ) = (
+                {', '.join(('?' for _ in columns))}
+            )
+        """
+        params = object_id + (tag,)
+
+        with self.db:
+            self.db.execute(query, params)
+
+    # Tuple[Optional[str], ...] seems hacky,
+    # but we can't have Optional[Tuple[str, ...]]
+    # because we depend on the tuple length
+    # to distinguish between feeds and entries.
+
+    def get_tags(self, object_id: Tuple[Optional[str], ...]) -> Iterable[str]:
         yield from join_paginated_iter(
-            partial(self.get_feed_tags_page, feed_url),
+            partial(self.get_tags_page, object_id),
             self.chunk_size,
         )
 
     @wrap_exceptions_iter(StorageError)
-    def get_feed_tags_page(
+    def get_tags_page(
         self,
-        feed_url: Optional[str],
+        object_id: Tuple[Optional[str], ...],
         chunk_size: Optional[int] = None,
         last: Optional[_T] = None,
     ) -> Iterable[Tuple[str, Optional[_T]]]:
-        query = Query().SELECT_DISTINCT("tag").FROM("feed_tags")
+        info = SCHEMA_INFO[len(object_id)]
+
+        query = Query().SELECT_DISTINCT("tag").FROM(f"{info.table_prefix}tags")
         context: Dict[str, Any] = dict()
 
-        if feed_url is not None:
-            query.WHERE("feed = :feed_url")
-            context.update(feed_url=feed_url)
+        if not any(p is None for p in object_id):
+            for column in info.id_columns:
+                query.WHERE(f"{column} = :{column}")
+            context.update(zip(info.id_columns, object_id))
 
         query.scrolling_window_order_by("tag")
 
-        def value_factory(t: Tuple[Any, ...]) -> str:
+        def row_factory(t: Tuple[Any, ...]) -> str:
             tag = t[0]
             assert isinstance(tag, str)
             return tag
 
         yield from paginated_query(
-            self.db, query, context, value_factory, chunk_size, last
+            self.db, query, context, chunk_size, last, row_factory
         )
 
 
+def make_get_feeds_query(
+    filter_options: FeedFilterOptions, sort: FeedSortOrder
+) -> Tuple[Query, Dict[str, Any]]:
+    query = (
+        Query()
+        .SELECT(
+            'url',
+            'updated',
+            'title',
+            'link',
+            'author',
+            'user_title',
+            'added',
+            'last_updated',
+            'last_exception',
+            'updates_enabled',
+        )
+        .FROM("feeds")
+    )
+
+    context = apply_feed_filter_options(query, filter_options)
+
+    # NOTE: when changing, ensure none of the values can be null
+    # to prevent https://github.com/lemon24/reader/issues/203
+
+    # sort by url at the end to make sure the order is deterministic
+    if sort == 'title':
+        query.SELECT(("kinda_title", "lower(coalesce(user_title, title, ''))"))
+        query.scrolling_window_order_by("kinda_title", "url")
+    elif sort == 'added':
+        query.SELECT("added")
+        query.scrolling_window_order_by("added", "url", desc=True)
+    else:
+        assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
+
+    return query, context
+
+
 def feed_factory(t: Tuple[Any, ...]) -> Feed:
     return Feed._make(
         t[:8]
@@ -1254,6 +1343,20 @@ def make_get_entries_query(
     return query, filter_context
 
 
+def entry_factory(t: Tuple[Any, ...]) -> Entry:
+    feed = feed_factory(t[0:10])
+    entry = t[10:17] + (
+        tuple(Content(**d) for d in json.loads(t[17])) if t[17] else (),
+        tuple(Enclosure(**d) for d in json.loads(t[18])) if t[18] else (),
+        t[19] == 1,
+        t[20] == 1,
+        t[21],
+        t[22] or feed.url,
+        feed,
+    )
+    return Entry._make(entry)
+
+
 def apply_entry_filter_options(
     query: Query, filter_options: EntryFilterOptions, keyword: str = 'WHERE'
 ) -> Dict[str, Any]:
@@ -1309,13 +1412,12 @@ def apply_feed_tags_filter_options(
     next_tag_id = 0
 
     for subtags in tags:
-        # TODO: Query() should allow building this kind of stuff.
-        or_parts = []
+        tag_query = BaseQuery({'(': [], ')': ['']}, {'(': 'OR'})
+        tag_add = partial(tag_query.add, '(')
+
         for maybe_tag in subtags:
             if isinstance(maybe_tag, bool):
-                or_parts.append(
-                    f"{'NOT' if not maybe_tag else ''} (SELECT * FROM tags_count)"
-                )
+                tag_add(f"{'NOT' if not maybe_tag else ''} (SELECT * FROM tags_count)")
                 add_tags_count_cte = True
                 continue
 
@@ -1323,10 +1425,10 @@ def apply_feed_tags_filter_options(
             tag_name = f'__tag_{next_tag_id}'
             next_tag_id += 1
             context[tag_name] = tag
-            or_parts.append(f":{tag_name} {'NOT' if is_negation else ''} IN tags")
+            tag_add(f":{tag_name} {'NOT' if is_negation else ''} IN tags")
             add_tags_cte = True
 
-        add(f"({' OR '.join(or_parts)})")
+        add(str(tag_query))
 
     if add_tags_cte:
         query.WITH(("tags", f"SELECT tag FROM feed_tags WHERE feed = {url_column}"))
@@ -1425,15 +1527,15 @@ def apply_recent(
         "negative_feed_order",
     )
 
+    # NOTE: when changing, ensure none of the values can be null
+    # to prevent https://github.com/lemon24/reader/issues/203
     query.scrolling_window_order_by(
-        *f"""
-        kinda_first_updated
-        kinda_published
-        {id_prefix}feed
-        ids.last_updated
-        negative_feed_order
-        {id_prefix}id
-        """.split(),
+        'kinda_first_updated',
+        'kinda_published',
+        f'{id_prefix}feed',
+        'ids.last_updated',
+        'negative_feed_order',
+        f'{id_prefix}id',
         desc=True,
         keyword=keyword,
     )
