@@ -1,14 +1,16 @@
 import builtins
-import datetime
 import itertools
 import logging
 import numbers
 import warnings
+from datetime import datetime
+from types import MappingProxyType
 from typing import Any
 from typing import Callable
 from typing import Collection
 from typing import Iterable
 from typing import Iterator
+from typing import Mapping
 from typing import Optional
 from typing import overload
 from typing import Tuple
@@ -18,23 +20,30 @@ from typing import Union
 import reader._updater
 from ._parser import default_parser
 from ._parser import Parser
+from ._parser import SESSION_TIMEOUT
+from ._requests_utils import TimeoutType
 from ._search import Search
 from ._storage import Storage
+from ._types import DEFAULT_RESERVED_NAME_SCHEME
 from ._types import EntryData
 from ._types import EntryFilterOptions
 from ._types import EntryUpdateIntent
 from ._types import FeedFilterOptions
 from ._types import FeedForUpdate
 from ._types import FeedUpdateIntent
+from ._types import NameScheme
 from ._types import ParsedFeed
-from ._utils import make_noop_context_manager
+from ._utils import deprecated_wrapper
 from ._utils import make_pool_map
+from ._utils import nullcontext
 from ._utils import zero_or_one
-from .exceptions import _NotModified
 from .exceptions import EntryNotFoundError
+from .exceptions import FeedMetadataNotFoundError
 from .exceptions import FeedNotFoundError
-from .exceptions import MetadataNotFoundError
+from .exceptions import InvalidPluginError
 from .exceptions import ParseError
+from .plugins import _DEFAULT_PLUGINS
+from .plugins import _PLUGINS
 from .types import _entry_argument
 from .types import _feed_argument
 from .types import Entry
@@ -52,6 +61,8 @@ from .types import MISSING
 from .types import MissingType
 from .types import SearchSortOrder
 from .types import TagFilterInput
+from .types import UpdatedFeed
+from .types import UpdateResult
 
 
 log = logging.getLogger('reader')
@@ -60,14 +71,18 @@ log = logging.getLogger('reader')
 _T = TypeVar('_T')
 _U = TypeVar('_U')
 
-
-_PostEntryAddPluginType = Callable[['Reader', EntryData[datetime.datetime]], None]
+ReaderPluginType = Callable[['Reader'], None]
+_PostFeedUpdatePluginType = Callable[['Reader', str], None]
+_PostEntryAddPluginType = Callable[['Reader', EntryData[datetime]], None]
 
 
 def make_reader(
     url: str,
     *,
     feed_root: Optional[str] = '',
+    plugins: Iterable[Union[str, ReaderPluginType]] = _DEFAULT_PLUGINS,
+    session_timeout: TimeoutType = SESSION_TIMEOUT,
+    reserved_name_scheme: Mapping[str, str] = DEFAULT_RESERVED_NAME_SCHEME,
     _storage: Optional[Storage] = None,
     _storage_factory: Any = None,
 ) -> 'Reader':
@@ -120,17 +135,40 @@ def make_reader(
 
     Args:
         url (str): Path to the reader database.
+
         feed_root (str or None):
             Directory where to look for local feeds.
             One of ``None`` (don't open local feeds),
             ``''`` (full filesystem access; default), or
             ``'/path/to/feed/root'`` (an absolute path that feed paths are relative to).
 
+        plugins (iterable(str or callable(Reader)) or None):
+            An iterable of built-in plugin names or
+            `plugin(reader) --> None` callables.
+            The callables are called with the reader object
+            before it is returned.
+            Exceptions from plugin code will propagate to the caller.
+            The only plugin used by default is ``reader.ua_fallback``.
+
+        session_timeout (float or tuple(float, float) or None):
+            When retrieving HTTP(S) feeds,
+            how many seconds to wait for the server to send data,
+            as a float, or a (connect timeout, read timeout) tuple.
+            Passed to the underlying `Requests session`_.
+
+        reserved_name_scheme (dict(str, str) or None):
+            Value for :attr:`~Reader.reserved_name_scheme`.
+            The prefixes default to ``.reader.``/``.plugin.``,
+            and the separator to ``.``
+
+    .. _Requests session: https://requests.readthedocs.io/en/master/user/advanced/#timeouts
+
     Returns:
         Reader: The reader.
 
     Raises:
         StorageError
+        InvalidPluginError: If an invalid plugin name is passed to ``plugins``.
 
     .. versionadded:: 1.6
         The ``feed_root`` keyword argument.
@@ -140,6 +178,18 @@ def make_reader(
         *full filesystem access* (``''``) to
         *don't open local feeds* (``None``).
 
+    .. versionadded:: 1.14
+        The ``session_timeout`` keyword argument,
+        with a default of (3.05, 60) seconds;
+        the previous behavior was to *never time out*.
+
+    .. versionadded:: 1.16
+        The ``plugins`` keyword argument. Using an invalid plugin name
+        raises :exc:`InvalidPluginError`, a :exc:`ValueError` subclass.
+
+    .. versionadded:: 1.17
+        The ``reserved_name_scheme`` argument.
+
     """
 
     # If we ever need to change the signature of make_reader(),
@@ -154,17 +204,32 @@ def make_reader(
     # For now, we're using a storage-bound search provider.
     search = Search(storage)
 
-    parser = default_parser(feed_root)
+    parser = default_parser(feed_root, session_timeout=session_timeout)
+
+    reader = Reader(
+        storage, search, parser, reserved_name_scheme, _called_directly=False
+    )
+
+    for plugin in plugins:
+        if isinstance(plugin, str):
+            if plugin not in _PLUGINS:
+                raise InvalidPluginError(f"no such built-in plugin: {plugin!r}")
+
+            plugin_func = _PLUGINS[plugin]
+        else:
+            plugin_func = plugin
+
+        try:
+            plugin_func(reader)  # type: ignore
+        except Exception:  # pragma: no cover
+            # TODO: this whole branch is not tested
+            reader.close()
+            # TODO: this should raise a custom exception (but can't because of backwards compatibility)
+            raise
 
-    reader = Reader(storage, search, parser, _called_directly=False)
     return reader
 
 
-# If we ever want to implement metrics for Reader, see this comment:
-# https://github.com/lemon24/reader/issues/68#issuecomment-450025175
-# TODO: gather all the design notes in one place
-
-
 class Reader:
 
     """A feed reader.
@@ -196,6 +261,10 @@ class Reader:
         If you have a strong use case preventing you to do so,
         please +1 / comment in :issue:`206`.
 
+
+    .. versionadded:: 1.13
+        JSON Feed support.
+
     """
 
     def __init__(
@@ -203,12 +272,20 @@ class Reader:
         _storage: Storage,
         _search: Search,
         _parser: Parser,
+        _reserved_name_scheme: Mapping[str, str],
         _called_directly: bool = True,
     ):
         self._storage = _storage
         self._search = _search
         self._parser = _parser
+
+        try:
+            self.reserved_name_scheme = _reserved_name_scheme
+        except AttributeError as e:
+            raise ValueError(str(e)) from (e.__cause__ or e)
+
         self._updater = reader._updater
+        self._post_feed_update_plugins: Collection[_PostFeedUpdatePluginType] = []
         self._post_entry_add_plugins: Collection[_PostEntryAddPluginType] = []
 
         if _called_directly:
@@ -248,10 +325,8 @@ class Reader:
         now = self._now()
         self._storage.add_feed(url, now)
 
-    def remove_feed(self, feed: FeedInput) -> None:
-        """Remove a feed.
-
-        Also removes all of the feed's entries, metadata, and tags.
+    def delete_feed(self, feed: FeedInput) -> None:
+        """Delete a feed and all of its entries, metadata, and tags.
 
         Args:
             feed (str or Feed): The feed URL.
@@ -260,9 +335,14 @@ class Reader:
             FeedNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`remove_feed`.
+
         """
         url = _feed_argument(feed)
-        self._storage.remove_feed(url)
+        self._storage.delete_feed(url)
+
+    remove_feed = deprecated_wrapper('remove_feed', delete_feed, '1.18', '2.0')
 
     def change_feed_url(self, old: FeedInput, new: FeedInput) -> None:
         """Change the URL of a feed.
@@ -462,7 +542,7 @@ class Reader:
                 Only count feeds that have updates enabled / disabled.
 
         Returns:
-            :class:`FeedCounts`:
+            FeedCounts:
 
         Raises:
             StorageError
@@ -531,6 +611,8 @@ class Reader:
 
         Silently skip feeds that raise :exc:`ParseError`.
 
+        Roughly equivalent to ``for _ in reader.update_feed_iter(...): pass``.
+
         Args:
             new_only (bool): Only update feeds that have never been updated.
             workers (int): Number of threads to use when getting the feeds.
@@ -541,37 +623,84 @@ class Reader:
         .. versionchanged:: 1.11
             Only update the feeds that have updates enabled.
 
-        """
+        .. versionchanged:: 1.15
+            Update entries whenever their content changes,
+            regardless of their :attr:`~Entry.updated` date.
 
+            Content-only updates (not due to an :attr:`~Entry.updated` change)
+            are limited to 24 consecutive updates,
+            to prevent spurious updates for entries whose content changes
+            excessively (for example, because it includes the current time).
+
+            Previously, entries would be updated only if the
+            entry :attr:`~Entry.updated` was *newer* than the stored one.
+
+        """
+        for url, value in self.update_feeds_iter(new_only, workers):
+            if isinstance(value, ParseError):
+                log.exception(
+                    "update feed %r: error while getting/parsing feed, "
+                    "skipping; exception: %r",
+                    url,
+                    value.__cause__,
+                    exc_info=value,
+                )
+                continue
+
+            assert not isinstance(value, Exception), value
+
+    def update_feeds_iter(
+        self, new_only: bool = False, workers: int = 1
+    ) -> Iterable[UpdateResult]:
+        """Update all the feeds that have updates enabled.
+
+        Args:
+            new_only (bool): Only update feeds that have never been updated.
+            workers (int): Number of threads to use when getting the feeds.
+
+        Yields:
+            :class:`UpdateResult`:
+                An (url, value) pair; the value is one of:
+
+                * a summary of the updated feed, if the update was successful
+                * None, if the server indicated the feed has not changed
+                  since the last update
+                * an exception instance
+
+                Currently, the exception is always a :exc:`ParseError`,
+                but other :exc:`ReaderError` subclasses may be yielded
+                in the future.
+
+        Raises:
+            StorageError
+
+        .. versionadded:: 1.14
+
+        .. versionchanged:: 1.15
+            Update entries whenever their content changes.
+            See :meth:`~Reader.update_feeds` for details.
+
+        """
         if workers < 1:
             raise ValueError("workers must be a positive integer")
 
-        make_map = (
-            make_noop_context_manager(builtins.map)
-            if workers == 1
-            else make_pool_map(workers)
-        )
+        make_map = nullcontext(builtins.map) if workers == 1 else make_pool_map(workers)
 
         with make_map as map:
-            exceptions = self._update_feeds(new_only=new_only, map=map)
+            results = self._update_feeds(new_only=new_only, map=map)
 
-            for exc in exceptions:
-                if not exc:
+            for url, value in results:
+                if isinstance(value, FeedNotFoundError):
+                    log.info("update feed %r: feed removed during update", url)
                     continue
-                if isinstance(exc, FeedNotFoundError):
-                    log.info("update feed %r: feed removed during update", exc.url)
-                elif isinstance(exc, ParseError):
-                    log.exception(
-                        "update feed %r: error while getting/parsing feed, "
-                        "skipping; exception: %r",
-                        exc.url,
-                        exc.__cause__,
-                        exc_info=exc,
-                    )
-                else:
-                    raise exc
 
-    def update_feed(self, feed: FeedInput) -> None:
+                if isinstance(value, Exception):
+                    if not isinstance(value, ParseError):
+                        raise value
+
+                yield UpdateResult(url, value)
+
+    def update_feed(self, feed: FeedInput) -> Optional[UpdatedFeed]:
         """Update a single feed.
 
         The feed will be updated even if updates are disabled for it.
@@ -579,23 +708,37 @@ class Reader:
         Args:
             feed (str or Feed): The feed URL.
 
+        Returns:
+            UpdatedFeed or None:
+            A summary of the updated feed or None,
+            if the server indicated the feed has not changed
+            since the last update.
+
         Raises:
             FeedNotFoundError
             ParseError
             StorageError
 
+        .. versionchanged:: 1.14
+            The method now returns UpdatedFeed or None instead of None.
+
+        .. versionchanged:: 1.15
+            Update entries whenever their content changes.
+            See :meth:`~Reader.update_feeds` for details.
+
         """
         url = _feed_argument(feed)
-        exc = zero_or_one(
+        _, rv = zero_or_one(
             self._update_feeds(url=url, enabled_only=False),
             lambda: FeedNotFoundError(url),
         )
-        if exc:
-            raise exc
+        if isinstance(rv, Exception):
+            raise rv
+        return rv
 
     @staticmethod
-    def _now() -> datetime.datetime:
-        return datetime.datetime.utcnow()
+    def _now() -> datetime:
+        return datetime.utcnow()
 
     # The type of map should be
     #
@@ -612,7 +755,7 @@ class Reader:
         new_only: bool = False,
         enabled_only: bool = True,
         map: Callable[[Callable[[Any], Any], Iterable[Any]], Iterator[Any]] = map,
-    ) -> Iterator[Optional[Exception]]:
+    ) -> Iterator[Tuple[str, Union[UpdatedFeed, None, Exception]]]:
 
         # global_now is used as first_updated_epoch for all new entries,
         # so that the subset of new entries from an update appears before
@@ -652,12 +795,13 @@ class Reader:
         pairs = map(self._parse_feed_for_update, feeds_for_update)
 
         for feed_for_update, parse_result in pairs:
+            rv: Union[UpdatedFeed, None, Exception]
 
             try:
                 # give storage a chance to consume the entries in a streaming fashion;
                 parsed_entries = itertools.tee(
                     parse_result.entries
-                    if not isinstance(parse_result, Exception)
+                    if parse_result and not isinstance(parse_result, Exception)
                     else ()
                 )
                 entry_pairs = zip(
@@ -671,22 +815,32 @@ class Reader:
                 (
                     feed_to_update,
                     entries_to_update,
-                    exception,
                 ) = self._updater.make_update_intents(
                     feed_for_update, now, global_now, parse_result, entry_pairs
                 )
 
-                self._update_feed(feed_to_update, entries_to_update)
-                yield exception
+                counts = self._update_feed(
+                    feed_for_update.url, feed_to_update, entries_to_update
+                )
+
+                if isinstance(parse_result, Exception):
+                    rv = parse_result
+                elif parse_result:
+                    rv = UpdatedFeed(feed_for_update.url, *counts)
+                else:
+                    rv = None
+
             except Exception as e:
-                yield e
+                rv = e
+
+            yield feed_for_update.url, rv
 
     def _parse_feed_for_update(
         self, feed: FeedForUpdate
-    ) -> Tuple[FeedForUpdate, Union[ParsedFeed, ParseError, _NotModified]]:
+    ) -> Tuple[FeedForUpdate, Union[ParsedFeed, None, ParseError]]:
         try:
             return feed, self._parser(feed.url, feed.http_etag, feed.http_last_modified)
-        except (ParseError, _NotModified) as e:
+        except ParseError as e:
             log.debug(
                 "_parse_feed_for_update exception, traceback follows", exc_info=True
             )
@@ -694,9 +848,10 @@ class Reader:
 
     def _update_feed(
         self,
+        url: str,
         feed_to_update: Optional[FeedUpdateIntent],
         entries_to_update: Iterable[EntryUpdateIntent],
-    ) -> None:
+    ) -> Tuple[int, int]:
         if feed_to_update:
             if entries_to_update:
                 self._storage.add_or_update_entries(entries_to_update)
@@ -705,11 +860,22 @@ class Reader:
         # if feed_for_update.url != parsed_feed.feed.url, the feed was redirected.
         # TODO: Maybe handle redirects somehow else (e.g. change URL if permanent).
 
+        for feed_plugin in self._post_feed_update_plugins:
+            feed_plugin(self, url)
+
+        new_count = 0
+        updated_count = 0
         for entry in entries_to_update:
+            if entry.new:
+                new_count += 1
+            else:
+                updated_count += 1
             if not entry.new:
                 continue
-            for plugin in self._post_entry_add_plugins:
-                plugin(self, entry.entry)
+            for entry_plugin in self._post_entry_add_plugins:
+                entry_plugin(self, entry.entry)
+
+        return new_count, updated_count
 
     def get_entries(
         self,
@@ -875,7 +1041,7 @@ class Reader:
                 Only count the entries from feeds matching these tags.
 
         Returns:
-            :class:`EntryCounts`:
+            EntryCounts:
 
         Raises:
             StorageError
@@ -889,7 +1055,7 @@ class Reader:
         )
         return self._storage.get_entry_counts(filter_options)
 
-    def mark_as_read(self, entry: EntryInput) -> None:
+    def mark_entry_as_read(self, entry: EntryInput) -> None:
         """Mark an entry as read.
 
         Args:
@@ -899,11 +1065,14 @@ class Reader:
             EntryNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`mark_as_read`.
+
         """
         feed_url, entry_id = _entry_argument(entry)
         self._storage.mark_as_read_unread(feed_url, entry_id, True)
 
-    def mark_as_unread(self, entry: EntryInput) -> None:
+    def mark_entry_as_unread(self, entry: EntryInput) -> None:
         """Mark an entry as unread.
 
         Args:
@@ -913,11 +1082,14 @@ class Reader:
             EntryNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`mark_as_unread`.
+
         """
         feed_url, entry_id = _entry_argument(entry)
         self._storage.mark_as_read_unread(feed_url, entry_id, False)
 
-    def mark_as_important(self, entry: EntryInput) -> None:
+    def mark_entry_as_important(self, entry: EntryInput) -> None:
         """Mark an entry as important.
 
         Args:
@@ -927,11 +1099,14 @@ class Reader:
             EntryNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`mark_as_important`.
+
         """
         feed_url, entry_id = _entry_argument(entry)
         self._storage.mark_as_important_unimportant(feed_url, entry_id, True)
 
-    def mark_as_unimportant(self, entry: EntryInput) -> None:
+    def mark_entry_as_unimportant(self, entry: EntryInput) -> None:
         """Mark an entry as unimportant.
 
         Args:
@@ -941,51 +1116,95 @@ class Reader:
             EntryNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`mark_as_unimportant`.
+
         """
         feed_url, entry_id = _entry_argument(entry)
         self._storage.mark_as_important_unimportant(feed_url, entry_id, False)
 
-    def iter_feed_metadata(
+    mark_as_read = deprecated_wrapper('mark_as_read', mark_entry_as_read, '1.18', '2.0')
+    mark_as_unread = deprecated_wrapper(
+        'mark_as_unread', mark_entry_as_unread, '1.18', '2.0'
+    )
+    mark_as_important = deprecated_wrapper(
+        'mark_as_important', mark_entry_as_important, '1.18', '2.0'
+    )
+    mark_as_unimportant = deprecated_wrapper(
+        'mark_as_unimportant', mark_entry_as_unimportant, '1.18', '2.0'
+    )
+
+    def get_feed_metadata(
         self,
         feed: FeedInput,
-        *,
+        *args: Any,
         key: Optional[str] = None,
     ) -> Iterable[Tuple[str, JSONType]]:
-        """Get all or some of the metadata values for a feed.
+        """Get all or some of the metadata for a feed as ``(key, value)`` pairs.
 
         Args:
             feed (str or Feed): The feed URL.
             key (str or None): Only return the metadata for this key.
 
         Yields:
-            tuple(str, JSONType): Key-value pairs, in undefined order.
+            tuple(str, JSONType): ``(key, value)`` pairs, in undefined order.
             JSONType is whatever :py:func:`json.dumps` accepts.
 
         Raises:
             StorageError
 
+        .. versionchanged:: 1.18
+
+            :meth:`iter_feed_metadata` was renamed to :meth:`get_feed_metadata`,
+            and :meth:`get_feed_metadata` was renamed to :meth:`get_feed_metadata_item`.
+
+            To preserve backwards compatibility,
+            the ``get_feed_metadata(feed, key[, default]) -> value``
+            form (positional arguments only)
+            will continue to work as an alias for
+            ``get_feed_metadata_item(feed, key[, default])``
+            until the last 1.\\* *reader* version,
+            after which it will result in a :exc:`TypeError`.
+
         """
+
+        if args:
+            # get_feed_metadata(feed, key[, default]) -> value
+            if len(args) > 2:
+                raise TypeError(
+                    f"get_feed_metadata() takes 1 positional arguments, but {len(args) + 1} were given"
+                )
+            warnings.warn(
+                "the get_feed_metadata(feed, key[, default]) -> value "
+                "version of get_feed_metadata() is deprecated "
+                "and will be removed in reader 2.0. "
+                "Use get_feed_metadata_item() instead.",
+                DeprecationWarning,
+            )
+            return self.get_feed_metadata_item(feed, *args)  # type: ignore
+
+        # get_feed_metadata(feed, *, key=None) -> (key, value), ...
         feed_url = _feed_argument(feed)
-        return self._storage.iter_feed_metadata(feed_url, key)
+        return self._storage.iter_metadata((feed_url,), key)
 
     @overload
-    def get_feed_metadata(
+    def get_feed_metadata_item(
         self, feed: FeedInput, key: str
     ) -> JSONType:  # pragma: no cover
         ...
 
     @overload
-    def get_feed_metadata(
+    def get_feed_metadata_item(
         self, feed: FeedInput, key: str, default: _T
     ) -> Union[JSONType, _T]:  # pragma: no cover
         ...
 
-    def get_feed_metadata(
+    def get_feed_metadata_item(
         self, feed: FeedInput, key: str, default: Union[MissingType, _T] = MISSING
     ) -> Union[JSONType, _T]:
         """Get metadata for a feed.
 
-        Like ``next(iter(reader.get_feed_metadata(feed, key=key)), default)``,
+        Like ``next(iter(reader.get_feed_metadata(feed, key=key)), (None, default))[1]``,
         but raises a custom exception instead of :exc:`StopIteration`.
 
         Args:
@@ -998,47 +1217,68 @@ class Reader:
             JSONType is whatever :py:func:`json.dumps` accepts.
 
         Raises:
-            MetadataNotFoundError
+            FeedMetadataNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`get_feed_metadata`.
+
         """
         return zero_or_one(
-            (v for _, v in self.iter_feed_metadata(feed, key=key)),
-            lambda: MetadataNotFoundError(_feed_argument(feed), key),
+            (v for _, v in self.get_feed_metadata(feed, key=key)),
+            lambda: FeedMetadataNotFoundError(_feed_argument(feed), key),
             default,
         )
 
-    def set_feed_metadata(self, feed: FeedInput, key: str, value: JSONType) -> None:
+    def set_feed_metadata_item(
+        self, feed: FeedInput, key: str, value: JSONType
+    ) -> None:
         """Set metadata for a feed.
 
         Args:
             feed (str or Feed): The feed URL.
-            key (str): The key of the metadata to set.
-            value (JSONType): The value of the metadata to set.
+            key (str): The key of the metadata item to set.
+            value (JSONType): The value of the metadata item to set.
                 JSONType is whatever :py:func:`json.dumps` accepts.
 
         Raises:
             FeedNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`set_feed_metadata`.
+
         """
         feed_url = _feed_argument(feed)
-        self._storage.set_feed_metadata(feed_url, key, value)
+        self._storage.set_metadata((feed_url,), key, value)
 
-    def delete_feed_metadata(self, feed: FeedInput, key: str) -> None:
+    def delete_feed_metadata_item(self, feed: FeedInput, key: str) -> None:
         """Delete metadata for a feed.
 
         Args:
             feed (str or Feed): The feed URL.
-            key (str): The key of the metadata to delete.
+            key (str): The key of the metadata item to delete.
 
         Raises:
-            MetadataNotFoundError
+            FeedMetadataNotFoundError
             StorageError
 
+        .. versionadded:: 1.18
+            Renamed from :meth:`delete_feed_metadata`.
+
         """
         feed_url = _feed_argument(feed)
-        self._storage.delete_feed_metadata(feed_url, key)
+        self._storage.delete_metadata((feed_url,), key)
+
+    iter_feed_metadata = deprecated_wrapper(
+        'iter_feed_metadata', get_feed_metadata, '1.18', '2.0'
+    )
+    set_feed_metadata = deprecated_wrapper(
+        'set_feed_metadata', set_feed_metadata_item, '1.18', '2.0'
+    )
+    delete_feed_metadata = deprecated_wrapper(
+        'delete_feed_metadata', delete_feed_metadata_item, '1.18', '2.0'
+    )
 
     def enable_search(self) -> None:
         """Enable full-text search.
@@ -1245,7 +1485,7 @@ class Reader:
                 Only count the entries from feeds matching these tags.
 
         Returns:
-            :class:`EntrySearchCounts`:
+            EntrySearchCounts:
 
         Raises:
             SearchNotEnabledError
@@ -1279,7 +1519,7 @@ class Reader:
 
         """
         feed_url = _feed_argument(feed)
-        self._storage.add_feed_tag(feed_url, tag)
+        self._storage.add_tag((feed_url,), tag)
 
     def remove_feed_tag(self, feed: FeedInput, tag: str) -> None:
         """Remove a tag from a feed.
@@ -1297,7 +1537,7 @@ class Reader:
 
         """
         feed_url = _feed_argument(feed)
-        self._storage.remove_feed_tag(feed_url, tag)
+        self._storage.remove_tag((feed_url,), tag)
 
     def get_feed_tags(self, feed: Optional[FeedInput] = None) -> Iterable[str]:
         """Get all or some of the feed tags.
@@ -1314,5 +1554,92 @@ class Reader:
         .. versionadded:: 1.7
 
         """
-        feed_url = _feed_argument(feed) if feed is not None else feed
-        return self._storage.get_feed_tags(feed_url)
+        feed_url = _feed_argument(feed) if feed is not None else None
+        return self._storage.get_tags((feed_url,))
+
+    def make_reader_reserved_name(self, key: str) -> str:
+        """Create a *reader*-reserved tag or metadata name.
+        See :ref:`reserved names` for details.
+
+        Uses :attr:`~Reader.reserved_name_scheme` to build names of the format::
+
+        {reader_prefix}{key}
+
+        Using the default scheme:
+
+        >>> reader.make_reader_reserved_name('key')
+        '.reader.key'
+
+        Args:
+            key (str): A key.
+
+        Returns:
+            str: The name.
+
+        .. versionadded:: 1.17
+
+        """
+        return self._reserved_name_scheme.make_reader_name(key)
+
+    def make_plugin_reserved_name(
+        self, plugin_name: str, key: Optional[str] = None
+    ) -> str:
+        """Create a plugin-reserved tag or metadata name.
+        See :ref:`reserved names` for details.
+
+        Plugins should use this to generate names
+        for plugin-specific tags and metadata.
+
+        Uses :attr:`~Reader.reserved_name_scheme` to build names of the format::
+
+        {plugin_prefix}{plugin_name}
+        {plugin_prefix}{plugin_name}{separator}{key}
+
+        Using the default scheme:
+
+        >>> reader.make_plugin_reserved_name('myplugin')
+        '.plugin.myplugin'
+        >>> reader.make_plugin_reserved_name('myplugin', 'key')
+        '.plugin.myplugin.key'
+
+        Args:
+            plugin_name (str): The plugin package/module name.
+            key (str or None): A key; if more than one reserved name is needed.
+
+        Returns:
+            str: The name.
+
+        .. versionadded:: 1.17
+
+        """
+        return self._reserved_name_scheme.make_plugin_name(plugin_name, key)
+
+    # Ideally, the getter would return a TypedDict,
+    # but the setter would take *any* Mapping[str, str];
+    # unfortunately, mypy doesn't like when the types differ:
+    # https://github.com/python/mypy/issues/3004
+
+    @property
+    def reserved_name_scheme(self) -> Mapping[str, str]:
+        """dict(str, str): Mapping used to build reserved names.
+        See :meth:`~Reader.make_reader_reserved_name`
+        and :meth:`~Reader.make_plugin_reserved_name`
+        for details on how this is used.
+
+        The default scheme (these keys are required)::
+
+            {'reader_prefix': '.reader.', 'plugin_prefix': '.plugin.', 'separator': '.'}
+
+        The returned mapping is immutable; assign a new mapping to change the scheme.
+
+        .. versionadded:: 1.17
+
+        """
+        return MappingProxyType(self._reserved_name_scheme.__dict__)
+
+    @reserved_name_scheme.setter
+    def reserved_name_scheme(self, value: Mapping[str, str]) -> None:
+        try:
+            self._reserved_name_scheme = NameScheme.from_value(value)
+        except Exception as e:
+            raise AttributeError(f"invalid scheme: {value}") from e
