@@ -11,6 +11,7 @@ from typing import Dict
 from typing import Iterable
 from typing import List
 from typing import Mapping
+from typing import NamedTuple
 from typing import Optional
 from typing import Sequence
 from typing import Tuple
@@ -22,6 +23,8 @@ from typing_extensions import Literal
 from typing_extensions import Protocol
 from typing_extensions import runtime_checkable
 
+from reader.exceptions import ReaderError
+
 
 _T = TypeVar('_T')
 
@@ -379,8 +382,9 @@ class HighlightedString:
         ['a', 'bc', 'd']
 
         Yields:
-            str: The parts. Parts with even indexes are highlighted,
-            parts with odd indexes are not.
+            str: The parts (always an odd number);
+            parts with odd indexes are highlighted,
+            parts with even indexes are not.
 
         """
         start = 0
@@ -621,3 +625,57 @@ class EntrySearchCounts(_namedtuple_compat):
 
     #: Number of entries that have enclosures.
     has_enclosures: Optional[int] = None
+
+
+@dataclass(frozen=True)
+class UpdatedFeed:
+    """The result of a successful feed update.
+
+    .. versionadded:: 1.14
+
+    """
+
+    #: The URL of the feed.
+    url: str
+
+    #: The number of new entries.
+    new: int
+
+    #: The number of updated entries.
+    updated: int
+
+
+class UpdateResult(NamedTuple):
+    """Named tuple representing the result of a feed update.
+
+    .. versionadded:: 1.14
+
+    """
+
+    #: The URL of the feed.
+    url: str
+
+    #: One of:
+    #:
+    #: :class:`UpdatedFeed`
+    #:
+    #:  If the update was successful; a summary of the updated feed.
+    #:
+    #: :obj:`None`
+    #:
+    #:  If the server indicated the feed has not changed
+    #:  since the last update.
+    #:
+    #: :exc:`ReaderError`
+    #:
+    #:  If there was an error while updating the feed.
+    #:
+    value: Union[UpdatedFeed, None, ReaderError]
+
+    # The exception type is ReaderError and not ParseError
+    # to allow suppressing new errors without breaking the API:
+    # adding a new type to the union breaks the API,
+    # not raising an exception type anymore doesn't.
+    # Currently, storage or plugin-raised exceptions
+    # prevent updates for the following feeds (:issue:`218`),
+    # but that's not necessarily by design.
