@@ -1,11 +1,13 @@
 import sqlite3
 import sys
+from contextlib import closing
+from functools import wraps
 
 import py.path
 import pytest
 from utils import reload_module
 
-from reader import make_reader
+from reader import make_reader as original_make_reader
 from reader._storage import Storage
 
 
@@ -45,14 +47,27 @@ def apply_flaky_pypy_sqlite3(items):  # pragma: no cover
         item.add_marker(sqlite3_flaky)
 
 
+@pytest.fixture
+def make_reader(request):
+    @wraps(original_make_reader)
+    def make_reader(*args, **kwargs):
+        reader = original_make_reader(*args, **kwargs)
+        request.addfinalizer(reader.close)
+        return reader
+
+    return make_reader
+
+
 @pytest.fixture
 def reader():
-    return make_reader(':memory:')
+    with closing(original_make_reader(':memory:')) as reader:
+        yield reader
 
 
 @pytest.fixture
 def storage():
-    return Storage(':memory:')
+    with closing(Storage(':memory:')) as storage:
+        yield storage
 
 
 def call_update_feeds(reader, _):
@@ -63,6 +78,16 @@ def call_update_feeds_workers(reader, _):
     reader.update_feeds(workers=2)
 
 
+def call_update_feeds_iter(reader, _):
+    for _ in reader.update_feeds_iter():
+        pass
+
+
+def call_update_feeds_iter_workers(reader, _):
+    for _ in reader.update_feeds_iter(workers=2):
+        pass
+
+
 def call_update_feed(reader, url):
     reader.update_feed(url)
 
@@ -71,6 +96,8 @@ def call_update_feed(reader, url):
     params=[
         call_update_feeds,
         pytest.param(call_update_feeds_workers, marks=pytest.mark.slow),
+        call_update_feeds_iter,
+        pytest.param(call_update_feeds_iter_workers, marks=pytest.mark.slow),
         call_update_feed,
     ]
 )
