@@ -1,4 +1,5 @@
 import logging
+import sys
 import threading
 from collections import Counter
 from datetime import datetime
@@ -7,13 +8,14 @@ from enum import Enum
 from itertools import permutations
 
 import pytest
-from fakeparser import _NotModifiedParser
 from fakeparser import BlockingParser
 from fakeparser import FailingParser
+from fakeparser import NotModifiedParser
 from fakeparser import Parser
 from utils import make_url_base
 from utils import rename_argument
 
+import reader._parser
 from reader import Content
 from reader import Enclosure
 from reader import Entry
@@ -23,121 +25,246 @@ from reader import EntrySearchCounts
 from reader import Feed
 from reader import FeedCounts
 from reader import FeedExistsError
+from reader import FeedMetadataNotFoundError
 from reader import FeedNotFoundError
-from reader import make_reader
-from reader import MetadataNotFoundError
+from reader import InvalidPluginError
 from reader import ParseError
 from reader import Reader
 from reader import StorageError
+from reader import UpdatedFeed
+from reader import UpdateResult
 from reader._storage import Storage
+from reader._types import DEFAULT_RESERVED_NAME_SCHEME
 from reader._types import FeedUpdateIntent
 
 
 # TODO: testing added/last_updated everywhere is kinda ugly
 
 
-def test_update_feed_updated(reader, call_update_method):
-    """If a feed is not newer than the stored one, it should not be updated,
-    but its entries should be processed anyway.
+def test_update_feed_updated(reader, call_update_method, caplog):
+    """If a feed is not newer than the stored one,
+    it should be updated only if its content (hash) changed.
+
+    Its entries should be processed anyway.
 
     Details in https://github.com/lemon24/reader/issues/76
 
     """
-
     parser = Parser()
     reader._parser = parser
 
+    # Initial update.
     old_feed = parser.feed(1, datetime(2010, 1, 1), title='old')
     entry_one = parser.entry(1, 1, datetime(2010, 1, 1))
 
     reader._now = lambda: datetime(2010, 1, 1)
     reader.add_feed(old_feed.url)
     reader._now = lambda: datetime(2010, 1, 2)
-    call_update_method(reader, old_feed.url)
-    feed = old_feed._replace(
+
+    with caplog.at_level(logging.DEBUG, 'reader'):
+        call_update_method(reader, old_feed.url)
+
+    feed = old_feed.as_feed(
         added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2)
     )
     assert set(reader.get_entries()) == {
         entry_one.as_entry(feed=feed, last_updated=datetime(2010, 1, 2))
     }
+    assert "feed has no last_updated, treating as updated" in caplog.text
+    caplog.clear()
 
-    parser.feed(1, datetime(2010, 1, 1), title='old-different-title')
+    # Entries should be processed anyway.
     entry_two = parser.entry(1, 2, datetime(2010, 2, 1))
     reader._now = lambda: datetime(2010, 1, 3)
-    call_update_method(reader, old_feed.url)
-    feed = old_feed._replace(
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, old_feed.url)
+
+    feed = old_feed.as_feed(
         added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 3)
     )
     assert set(reader.get_entries()) == {
         entry_one.as_entry(feed=feed, last_updated=datetime(2010, 1, 2)),
         entry_two.as_entry(feed=feed, last_updated=datetime(2010, 1, 3)),
     }
+    assert "feed not updated, updating entries anyway" in caplog.text
+    caplog.clear()
 
-    even_older_feed = parser.feed(1, datetime(2009, 1, 1), title='even-older')
+    # Feed gets updated because content (hash) changed.
+    old_feed = parser.feed(1, datetime(2010, 1, 1), title='old-different-title')
+    reader._now = lambda: datetime(2010, 1, 3, 12)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, old_feed.url)
+
+    feed = old_feed.as_feed(
+        added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 3, 12)
+    )
+    assert reader.get_feed(feed) == feed
+    assert "feed hash changed, treating as updated" in caplog.text
+    caplog.clear()
+
+    # The feed doesn't change, because .updated is older.
+    # Entries get updated regardless.
+    old_feed = parser.feed(1, datetime(2009, 1, 1), title='old-different-title')
     entry_three = parser.entry(1, 3, datetime(2010, 2, 1))
     reader._now = lambda: datetime(2010, 1, 4)
-    call_update_method(reader, old_feed.url)
-    feed = old_feed._replace(
-        added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 4)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, old_feed.url)
+
+    feed = old_feed.as_feed(
+        added=datetime(2010, 1, 1),
+        # doesn't change because it's not newer
+        updated=datetime(2010, 1, 1),
+        # changes because entries changed
+        last_updated=datetime(2010, 1, 4),
     )
     assert set(reader.get_entries()) == {
         entry_one.as_entry(feed=feed, last_updated=datetime(2010, 1, 2)),
         entry_two.as_entry(feed=feed, last_updated=datetime(2010, 1, 3)),
         entry_three.as_entry(feed=feed, last_updated=datetime(2010, 1, 4)),
     }
+    assert "feed not updated, updating entries anyway" in caplog.text
+    caplog.clear()
 
+    # The feed doesn't change; despite being newer, no entries have changed.
+    old_feed = parser.feed(1, datetime(2010, 1, 2), title='old-different-title')
+    reader._now = lambda: datetime(2010, 1, 4, 12)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, old_feed.url)
+
+    feed = old_feed.as_feed(
+        added=datetime(2010, 1, 1),
+        # doesn't change because no entries have changed
+        updated=datetime(2010, 1, 1),
+        # doesn't change because nothing changed
+        last_updated=datetime(2010, 1, 4),
+    )
+    assert set(reader.get_entries()) == {
+        entry_one.as_entry(feed=feed, last_updated=datetime(2010, 1, 2)),
+        entry_two.as_entry(feed=feed, last_updated=datetime(2010, 1, 3)),
+        entry_three.as_entry(feed=feed, last_updated=datetime(2010, 1, 4)),
+    }
+    assert "feed not updated, updating entries anyway" in caplog.text
+    caplog.clear()
+
+    # The feeds changes because it is newer *and* entries get updated.
     new_feed = parser.feed(1, datetime(2010, 1, 2), title='new')
     entry_four = parser.entry(1, 4, datetime(2010, 2, 1))
     reader._now = lambda: datetime(2010, 1, 5)
-    feed = new_feed._replace(
+    feed = new_feed.as_feed(
         added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 5)
     )
-    call_update_method(reader, old_feed.url)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, old_feed.url)
+
     assert set(reader.get_entries()) == {
         entry_one.as_entry(feed=feed, last_updated=datetime(2010, 1, 2)),
         entry_two.as_entry(feed=feed, last_updated=datetime(2010, 1, 3)),
         entry_three.as_entry(feed=feed, last_updated=datetime(2010, 1, 4)),
         entry_four.as_entry(feed=feed, last_updated=datetime(2010, 1, 5)),
     }
+    assert "feed updated" in caplog.text
+    caplog.clear()
 
 
-def test_update_entry_updated(reader, call_update_method):
-    """An entry should be updated only if it is newer than the stored one."""
+def test_update_entry_updated(reader, call_update_method, caplog, monkeypatch):
+    """An entry should be updated only if
+    it is newer than the stored one OR its content (hash) changed.
 
+    """
     parser = Parser()
     reader._parser = parser
 
+    # Initial update.
     feed = parser.feed(1, datetime(2010, 1, 1))
     old_entry = parser.entry(1, 1, datetime(2010, 1, 1))
 
     reader._now = lambda: datetime(2010, 2, 1)
     reader.add_feed(feed.url)
     reader._now = lambda: datetime(2010, 2, 2)
-    call_update_method(reader, feed.url)
-    feed = feed._replace(added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 2))
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, feed.url)
+
+    feed = feed.as_feed(added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 2))
     assert set(reader.get_entries()) == {
         old_entry.as_entry(feed=feed, last_updated=datetime(2010, 2, 2))
     }
+    assert "entry updated" in caplog.text
+    caplog.clear()
 
+    # Feed newer (doesn't change), entry remains unchanged.
+    feed = parser.feed(1, datetime(2010, 1, 2))
+    reader._now = lambda: datetime(2010, 2, 3)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, feed.url)
+
+    feed = feed.as_feed(
+        added=datetime(2010, 2, 1),
+        updated=datetime(2010, 1, 1),
+        last_updated=datetime(2010, 2, 2),
+    )
+    assert set(reader.get_entries()) == {
+        old_entry.as_entry(feed=feed, last_updated=datetime(2010, 2, 2))
+    }
+    assert "entry not updated, skipping" in caplog.text
+    assert "entry hash changed, updating" not in caplog.text
+    caplog.clear()
+
+    # Feed does not change, entry hash changes.
     feed = parser.feed(1, datetime(2010, 1, 2))
     new_entry = old_entry._replace(title='New Entry')
     parser.entries[1][1] = new_entry
-    reader._now = lambda: datetime(2010, 2, 3)
-    call_update_method(reader, feed.url)
-    feed = feed._replace(added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 3))
-    assert set(reader.get_entries()) == {
-        old_entry.as_entry(feed=feed, last_updated=datetime(2010, 2, 2))
-    }
+    reader._now = lambda: datetime(2010, 2, 3, 12)
 
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, feed.url)
+
+    feed = feed.as_feed(
+        added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 3, 12)
+    )
+    assert set(reader.get_entries()) == {
+        new_entry.as_entry(feed=feed, last_updated=datetime(2010, 2, 3, 12))
+    }
+    assert "entry not updated, skipping" in caplog.text
+    assert "entry hash changed, updating" in caplog.text
+    caplog.clear()
+
+    # Entry is newer.
     feed = parser.feed(1, datetime(2010, 1, 3))
     new_entry = new_entry._replace(updated=datetime(2010, 1, 2))
     parser.entries[1][1] = new_entry
     reader._now = lambda: datetime(2010, 2, 4)
-    call_update_method(reader, feed.url)
-    feed = feed._replace(added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 4))
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        call_update_method(reader, feed.url)
+
+    feed = feed.as_feed(added=datetime(2010, 2, 1), last_updated=datetime(2010, 2, 4))
     assert set(reader.get_entries()) == {
         new_entry.as_entry(feed=feed, last_updated=datetime(2010, 2, 4))
     }
+    assert "entry updated" in caplog.text
+    caplog.clear()
+
+    # Entry hash changes, but reaches the update limit.
+    reader._now = lambda: datetime(2010, 2, 5)
+    monkeypatch.setattr('reader._updater.HASH_CHANGED_LIMIT', 3)
+
+    with caplog.at_level(logging.DEBUG, logger='reader'):
+        for i in range(1, 6):
+            new_entry = new_entry._replace(title=f"Even Newer: change #{i}")
+            parser.entries[1][1] = new_entry
+            call_update_method(reader, feed.url)
+
+    assert set(e.title for e in reader.get_entries()) == {"Even Newer: change #3"}
+    assert caplog.text.count("entry hash changed, updating") == 3
+    assert caplog.text.count("entry hash changed, but exceeds the update limit") == 2
+    caplog.clear()
 
 
 @pytest.mark.parametrize('chunk_size', [Storage.chunk_size, 1])
@@ -165,7 +292,7 @@ def test_update_no_updated(reader, chunk_size, call_update_method):
     reader._now = lambda: datetime(2010, 1, 1)
     reader.add_feed(feed.url)
     call_update_method(reader, feed)
-    feed = feed._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
+    feed = feed.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
 
     assert set(reader.get_feeds()) == {feed}
     assert set(reader.get_entries()) == {
@@ -179,7 +306,7 @@ def test_update_no_updated(reader, chunk_size, call_update_method):
     entry_two = parser.entry(1, 2, None)
     reader._now = lambda: datetime(2010, 1, 2)
     call_update_method(reader, feed)
-    feed = feed._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
+    feed = feed.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
 
     assert set(reader.get_feeds()) == {feed}
     assert set(reader.get_entries()) == {
@@ -193,7 +320,7 @@ def test_update_no_updated(reader, chunk_size, call_update_method):
 
 
 @pytest.mark.slow
-def test_update_blocking(db_path, call_update_method):
+def test_update_blocking(db_path, make_reader, call_update_method):
     """Calls to reader._parser() shouldn't block the underlying storage."""
 
     parser = Parser()
@@ -211,9 +338,15 @@ def test_update_blocking(db_path, call_update_method):
     blocking_parser = BlockingParser.from_parser(parser)
 
     def target():
+        # can't use fixture because it would run close() in a different thread
+        from reader import make_reader
+
         reader = make_reader(db_path)
         reader._parser = blocking_parser
-        call_update_method(reader, feed.url)
+        try:
+            call_update_method(reader, feed.url)
+        finally:
+            reader.close()
 
     t = threading.Thread(target=target)
     t.start()
@@ -222,7 +355,7 @@ def test_update_blocking(db_path, call_update_method):
 
     try:
         # shouldn't raise an exception
-        reader.mark_as_read((feed.url, entry.id))
+        reader.mark_entry_as_read((feed.url, entry.id))
     finally:
         blocking_parser.can_return_from_parser.set()
         t.join()
@@ -241,7 +374,7 @@ def test_update_not_modified(reader, call_update_method):
     parser.feed(1, datetime(2010, 1, 2))
     parser.entry(1, 1, datetime(2010, 1, 2))
 
-    not_modified_parser = _NotModifiedParser.from_parser(parser)
+    not_modified_parser = NotModifiedParser.from_parser(parser)
     reader._parser = not_modified_parser
 
     # shouldn't raise an exception
@@ -269,7 +402,7 @@ def test_update_new_only(reader):
     reader.add_feed(two.url)
     reader.update_feeds(new_only=True)
 
-    two = two._replace(added=datetime(2010, 1, 2), last_updated=datetime(2010, 1, 2))
+    two = two.as_feed(added=datetime(2010, 1, 2), last_updated=datetime(2010, 1, 2))
     assert len(set(reader.get_feeds())) == 2
     assert set(reader.get_entries()) == {
         entry_two.as_entry(feed=two, last_updated=datetime(2010, 1, 2))
@@ -278,7 +411,7 @@ def test_update_new_only(reader):
     reader._now = lambda: datetime(2010, 1, 3)
     reader.update_feeds()
 
-    one = one._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 3))
+    one = one.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 3))
     assert len(set(reader.get_feeds())) == 2
     assert set(reader.get_entries()) == {
         entry_one.as_entry(feed=one, last_updated=datetime(2010, 1, 3)),
@@ -318,7 +451,7 @@ def test_update_new_only_not_modified(reader):
     https://github.com/lemon24/reader/issues/95
 
     """
-    parser = _NotModifiedParser()
+    parser = NotModifiedParser()
     reader._parser = parser
 
     feed = parser.feed(1, datetime(2010, 1, 1))
@@ -501,7 +634,7 @@ def updated_feeds_parser(parser):
 
 
 def raises_not_modified_parser(_):
-    return _NotModifiedParser()
+    return NotModifiedParser()
 
 
 @pytest.mark.parametrize(
@@ -573,7 +706,9 @@ class EntryAction(Enum):
         }
     ],
 )
-def test_update_feed_deleted(db_path, call_update_method, feed_action, entry_action):
+def test_update_feed_deleted(
+    db_path, make_reader, call_update_method, feed_action, entry_action
+):
     """reader.update_feed should raise FeedNotFoundError if the feed is
     deleted during parsing.
 
@@ -596,7 +731,7 @@ def test_update_feed_deleted(db_path, call_update_method, feed_action, entry_act
             parser.entry(1, 1, datetime(2010, 1, 2))
 
     if feed_action is FeedAction.update:
-        feed = parser.feed(1, datetime(2010, 1, 2))
+        feed = parser.feed(1, datetime(2010, 1, 2), title='new title')
 
     if feed_action is not FeedAction.fail:
         parser_cls = BlockingParser
@@ -608,12 +743,16 @@ def test_update_feed_deleted(db_path, call_update_method, feed_action, entry_act
     blocking_parser = parser_cls.from_parser(parser)
 
     def target():
+        # can't use fixture because it would run close() in a different thread
+        from reader import make_reader
+
+        blocking_parser.in_parser.wait()
+        reader = make_reader(db_path)
         try:
-            blocking_parser.in_parser.wait()
-            reader = make_reader(db_path)
-            reader.remove_feed(feed.url)
+            reader.delete_feed(feed.url)
         finally:
             blocking_parser.can_return_from_parser.set()
+            reader.close()
 
     t = threading.Thread(target=target)
     t.start()
@@ -647,8 +786,8 @@ def test_update_feed(reader, feed_arg):
         reader.update_feed(feed_arg(one))
 
     reader._now = lambda: datetime(2010, 1, 1)
-    one = one._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
-    two = two._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
+    one = one.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
+    two = two.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 1))
 
     reader.add_feed(one.url)
     reader.add_feed(two.url)
@@ -677,6 +816,102 @@ def test_update_feed(reader, feed_arg):
         reader.update_feed(feed_arg(one))
 
 
+def call_update_feeds_iter(reader):
+    yield from reader.update_feeds_iter()
+
+
+def call_update_feed_iter(reader):
+    for feed in reader.get_feeds(updates_enabled=True):
+        try:
+            yield feed.url, reader.update_feed(feed)
+        except ParseError as e:
+            yield feed.url, e
+
+
+@pytest.fixture(
+    params=[
+        call_update_feeds_iter,
+        call_update_feed_iter,
+    ]
+)
+def call_update_iter_method(request):
+    return request.param
+
+
+def test_update_feeds_iter(reader, call_update_iter_method):
+    reader._parser = parser = FailingParser(condition=lambda url: url == '3')
+
+    one = parser.feed(1, datetime(2010, 1, 1))
+    one_one = parser.entry(1, 1, datetime(2010, 1, 1))
+    one_two = parser.entry(1, 2, datetime(2010, 2, 1))
+    two = parser.feed(2, datetime(2010, 1, 1))
+    two_one = parser.entry(2, 1, datetime(2010, 2, 1))
+
+    for feed in one, two:
+        reader.add_feed(feed)
+
+    assert dict(call_update_iter_method(reader)) == {
+        '1': UpdatedFeed(url='1', new=2, updated=0),
+        '2': UpdatedFeed(url='2', new=1, updated=0),
+    }
+
+    assert next(call_update_iter_method(reader)) == UpdateResult(
+        '1', UpdatedFeed(url='1', new=0, updated=0)
+    )
+
+    one_two = parser.entry(1, 2, datetime(2010, 2, 2), title='new title')
+    one_three = parser.entry(1, 3, datetime(2010, 2, 1))
+    one_four = parser.entry(1, 4, datetime(2010, 2, 1))
+    three = parser.feed(3, datetime(2010, 1, 1))
+
+    reader.add_feed(three)
+
+    rv = dict(call_update_iter_method(reader))
+    assert set(rv) == set('123')
+
+    assert rv['1'] == UpdatedFeed(url='1', new=2, updated=1)
+    assert rv['2'] == UpdatedFeed(url='2', new=0, updated=0)
+
+    assert isinstance(rv['3'], ParseError)
+    assert rv['3'].url == '3'
+    assert rv['3'].__cause__ is parser.exception
+
+    reader._parser = parser = NotModifiedParser()
+
+    assert dict(call_update_iter_method(reader)) == dict.fromkeys('123')
+
+
+@pytest.mark.parametrize('exc_type', [StorageError, Exception])
+def test_update_feeds_iter_raised_exception(reader, exc_type, call_update_iter_method):
+    reader._parser = parser = Parser()
+
+    one = parser.feed(1, datetime(2010, 1, 1))
+    two = parser.feed(2, datetime(2010, 1, 1))
+    three = parser.feed(3, datetime(2010, 1, 1))
+
+    for feed in one, two, three:
+        reader.add_feed(feed)
+
+    original_storage_update_feed = reader._storage.update_feed
+
+    def storage_update_feed(intent):
+        if intent.url == '2':
+            raise exc_type('message')
+        return original_storage_update_feed(intent)
+
+    reader._storage.update_feed = storage_update_feed
+
+    rv = {}
+    with pytest.raises(exc_type) as excinfo:
+        rv.update(call_update_iter_method(reader))
+    assert 'message' in str(excinfo.value)
+
+    if not sys.implementation.name == 'pypy':
+        # for some reason, on PyPy the updates sometimes
+        # happen out of order and rv is empty
+        assert rv == {'1': UpdatedFeed(url='1', new=0, updated=0)}
+
+
 def test_mark_as_read_unread(reader, entry_arg):
     # TODO: Test read/unread the same way important/unimportant are (or the other way around).
 
@@ -689,41 +924,41 @@ def test_mark_as_read_unread(reader, entry_arg):
     entry_with_feed = entry.as_entry(feed=feed)
 
     with pytest.raises(EntryNotFoundError) as excinfo:
-        reader.mark_as_read(entry_arg(entry_with_feed))
-    assert (excinfo.value.url, excinfo.value.id) == (entry.feed_url, entry.id)
+        reader.mark_entry_as_read(entry_arg(entry_with_feed))
+    assert (excinfo.value.feed_url, excinfo.value.id) == (entry.feed_url, entry.id)
     assert 'no such entry' in excinfo.value.message
 
     with pytest.raises(EntryNotFoundError) as excinfo:
-        reader.mark_as_unread(entry_arg(entry_with_feed))
-    assert (excinfo.value.url, excinfo.value.id) == (entry.feed_url, entry.id)
+        reader.mark_entry_as_unread(entry_arg(entry_with_feed))
+    assert (excinfo.value.feed_url, excinfo.value.id) == (entry.feed_url, entry.id)
     assert 'no such entry' in excinfo.value.message
 
     reader.add_feed(feed.url)
 
     with pytest.raises(EntryNotFoundError):
-        reader.mark_as_read(entry_arg(entry_with_feed))
+        reader.mark_entry_as_read(entry_arg(entry_with_feed))
 
     with pytest.raises(EntryNotFoundError):
-        reader.mark_as_unread(entry_arg(entry_with_feed))
+        reader.mark_entry_as_unread(entry_arg(entry_with_feed))
 
     reader.update_feeds()
 
     (entry,) = list(reader.get_entries())
     assert not entry.read
 
-    reader.mark_as_read(entry_arg(entry_with_feed))
+    reader.mark_entry_as_read(entry_arg(entry_with_feed))
     (entry,) = list(reader.get_entries())
     assert entry.read
 
-    reader.mark_as_read(entry_arg(entry_with_feed))
+    reader.mark_entry_as_read(entry_arg(entry_with_feed))
     (entry,) = list(reader.get_entries())
     assert entry.read
 
-    reader.mark_as_unread(entry_arg(entry_with_feed))
+    reader.mark_entry_as_unread(entry_arg(entry_with_feed))
     (entry,) = list(reader.get_entries())
     assert not entry.read
 
-    reader.mark_as_unread(entry_arg(entry_with_feed))
+    reader.mark_entry_as_unread(entry_arg(entry_with_feed))
     (entry,) = list(reader.get_entries())
     assert not entry.read
 
@@ -978,6 +1213,12 @@ def test_get_entries_recent_feed_order(reader, chunk_size, pre_stuff, call_metho
     assert [eval(e.id)[1] for e in call_method(reader)] == [1, 4, 2, 3]
 
 
+# sqlite3 on PyPy can be brittle
+# (spurious "InterfaceError: Error binding parameter X")
+# and we're doing lots of tiny queries here which may trigger it,
+# so don't bother
+@pytest.mark.skipif("sys.implementation.name == 'pypy'")
+@pytest.mark.slow
 @pytest.mark.parametrize('chunk_size', [1, 2, 3, 4])
 @pytest.mark.parametrize(
     'pre_stuff, call_method',
@@ -1067,11 +1308,11 @@ def test_add_remove_get_feeds(reader, feed_arg):
     assert 'no such feed' in excinfo.value.message
 
     assert reader.get_feed(feed_arg(one), None) == None
-    assert reader.get_feed(feed_arg(one), default=1) == 1
+    assert reader.get_feed(feed_arg(one), 1) == 1
     assert set(reader.get_entries()) == set()
 
     with pytest.raises(FeedNotFoundError) as excinfo:
-        reader.remove_feed(feed_arg(one))
+        reader.delete_feed(feed_arg(one))
     assert excinfo.value.url == one.url
     assert 'no such feed' in excinfo.value.message
 
@@ -1093,8 +1334,8 @@ def test_add_remove_get_feeds(reader, feed_arg):
     reader._now = lambda: datetime(2010, 1, 2)
     reader.update_feeds()
 
-    one = one._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
-    two = two._replace(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
+    one = one.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
+    two = two.as_feed(added=datetime(2010, 1, 1), last_updated=datetime(2010, 1, 2))
     entry_one = entry_one.as_entry(feed=one, last_updated=datetime(2010, 1, 2))
     entry_two = entry_two.as_entry(feed=two, last_updated=datetime(2010, 1, 2))
 
@@ -1102,13 +1343,13 @@ def test_add_remove_get_feeds(reader, feed_arg):
     assert reader.get_feed(feed_arg(one)) == one
     assert set(reader.get_entries()) == {entry_one, entry_two}
 
-    reader.remove_feed(feed_arg(one))
+    reader.delete_feed(feed_arg(one))
     assert set(reader.get_feeds()) == {two}
     assert reader.get_feed(feed_arg(one), None) == None
     assert set(reader.get_entries()) == {entry_two}
 
     with pytest.raises(FeedNotFoundError) as excinfo:
-        reader.remove_feed(feed_arg(one))
+        reader.delete_feed(feed_arg(one))
 
 
 def test_get_feeds_sort_error(reader):
@@ -1116,7 +1357,8 @@ def test_get_feeds_sort_error(reader):
         set(reader.get_feeds(sort='bad sort'))
 
 
-def test_get_feeds_order_title(reader):
+@pytest.mark.parametrize('chunk_size', [Storage.chunk_size, 1, 2, 0])
+def test_get_feeds_order_title(reader, chunk_size):
     """When sort='title', feeds should be sorted by (with decreasing
     priority):
 
@@ -1127,6 +1369,9 @@ def test_get_feeds_order_title(reader):
     https://github.com/lemon24/reader/issues/102
 
     """
+    # for https://github.com/lemon24/reader/issues/203
+    reader._storage.chunk_size = chunk_size
+
     parser = Parser()
     reader._parser = parser
 
@@ -1183,7 +1428,6 @@ def test_get_feeds_order_added(reader):
     https://github.com/lemon24/reader/issues/98
 
     """
-
     parser = Parser()
     reader._parser = parser
 
@@ -1264,13 +1508,52 @@ def test_data_roundtrip(reader):
 
     assert list(reader.get_entries()) == [
         entry.as_entry(
-            feed=feed._replace(
+            feed=feed.as_feed(
                 added=datetime(2010, 1, 2), last_updated=datetime(2010, 1, 3)
             ),
             last_updated=datetime(2010, 1, 3),
         )
     ]
 
+    # TODO: this should be a different test
+    (feed_for_update,) = reader._storage.get_feeds_for_update(url=feed.url)
+    assert feed.hash == feed_for_update.hash
+    (entry_for_update,) = reader._storage.get_entries_for_update(
+        [(entry.feed_url, entry.id)]
+    )
+    assert entry.hash == entry_for_update.hash
+
+
+def test_data_hashes_remain_stable():
+    # TODO: note the duplication from test_data_roundtrip()
+
+    parser = Parser()
+    feed = parser.feed(1, datetime(2010, 1, 1), author='feed author')
+    entry = parser.entry(
+        1,
+        1,
+        datetime(2010, 1, 1),
+        author='entry author',
+        summary='summary',
+        content=(Content('value3', 'type', 'en'), Content('value2')),
+        enclosures=(Enclosure('http://e1', 'type', 1000), Enclosure('http://e2')),
+    )
+
+    assert feed.hash == b'\x00\xda\xf5\xa1Je\x13],\xf0\xdb\xaa\x88d\x99\xc6'
+    assert entry.hash == b'\x00f\xa9\xdb\t5\xdf\xedcK\xd9bm\x80,l'
+
+    assert feed._replace(url='x', updated='x').hash == feed.hash
+    assert (
+        feed._replace(title='x').hash
+        == b'\x00\xce\x81\xc7\x8d(\xab\xd8)\x06\x90?\xf9\x847\xc4'
+    )
+
+    assert entry._replace(feed_url='x', id='x', updated='x').hash == entry.hash
+    assert (
+        entry._replace(title='x').hash
+        == b'\x00\x95\xc4\xe9\xd3\x95\xf6\xff\xf0*\xbd\x00L\x08\x1a\xa2'
+    )
+
 
 @pytest.mark.parametrize('feed_type', ['rss', 'atom'])
 def test_integration(reader, feed_type, data_dir, monkeypatch):
@@ -1292,7 +1575,8 @@ def test_integration(reader, feed_type, data_dir, monkeypatch):
         class datetime_mock(datetime):
             pass
 
-        monkeypatch.setattr('datetime.datetime', datetime_mock)
+        # reader.core must "from datetime import datetime" !
+        monkeypatch.setattr('reader.core.datetime', datetime_mock)
 
     monkeypatch.setattr(datetime_mock, 'utcnow', lambda: datetime(2010, 1, 1))
     reader.add_feed(feed_url)
@@ -1318,38 +1602,38 @@ def test_integration(reader, feed_type, data_dir, monkeypatch):
 
 def test_feed_metadata(reader):
     with pytest.raises(FeedNotFoundError) as excinfo:
-        reader.set_feed_metadata('one', 'key', 'value')
+        reader.set_feed_metadata_item('one', 'key', 'value')
     assert excinfo.value.url == 'one'
     assert 'no such feed' in excinfo.value.message
 
-    with pytest.raises(MetadataNotFoundError) as excinfo:
-        reader.delete_feed_metadata('one', 'key')
+    with pytest.raises(FeedMetadataNotFoundError) as excinfo:
+        reader.delete_feed_metadata_item('one', 'key')
     assert (excinfo.value.url, excinfo.value.key) == ('one', 'key')
     assert 'no such metadata' in excinfo.value.message
 
     reader.add_feed('feed')
 
-    assert set(reader.iter_feed_metadata('feed')) == set()
-    with pytest.raises(MetadataNotFoundError) as excinfo:
-        reader.get_feed_metadata('feed', 'key')
+    assert set(reader.get_feed_metadata('feed')) == set()
+    with pytest.raises(FeedMetadataNotFoundError) as excinfo:
+        reader.get_feed_metadata_item('feed', 'key')
     assert (excinfo.value.url, excinfo.value.key) == ('feed', 'key')
     assert 'no such metadata' in excinfo.value.message
-    assert reader.get_feed_metadata('feed', 'key', None) is None
-    assert reader.get_feed_metadata('feed', 'key', default=0) == 0
+    assert reader.get_feed_metadata_item('feed', 'key', None) is None
+    assert reader.get_feed_metadata_item('feed', 'key', 0) == 0
 
-    with pytest.raises(MetadataNotFoundError):
-        reader.delete_feed_metadata('one', 'key')
+    with pytest.raises(FeedMetadataNotFoundError):
+        reader.delete_feed_metadata_item('one', 'key')
 
-    reader.set_feed_metadata('feed', 'key', 'value')
+    reader.set_feed_metadata_item('feed', 'key', 'value')
 
-    assert set(reader.iter_feed_metadata('feed')) == {('key', 'value')}
-    assert reader.get_feed_metadata('feed', 'key') == 'value'
+    assert set(reader.get_feed_metadata('feed')) == {('key', 'value')}
+    assert reader.get_feed_metadata_item('feed', 'key') == 'value'
 
-    reader.delete_feed_metadata('feed', 'key')
+    reader.delete_feed_metadata_item('feed', 'key')
 
-    assert set(reader.iter_feed_metadata('feed')) == set()
-    with pytest.raises(MetadataNotFoundError):
-        reader.get_feed_metadata('feed', 'key')
+    assert set(reader.get_feed_metadata('feed')) == set()
+    with pytest.raises(FeedMetadataNotFoundError):
+        reader.get_feed_metadata_item('feed', 'key')
 
 
 def test_get_entry(reader, entry_arg):
@@ -1364,10 +1648,10 @@ def test_get_entry(reader, entry_arg):
 
     with pytest.raises(EntryNotFoundError) as excinfo:
         reader.get_entry(entry_arg(entry.as_entry(feed=feed)))
-    assert (excinfo.value.url, excinfo.value.id) == (entry.feed_url, entry.id)
+    assert (excinfo.value.feed_url, excinfo.value.id) == (entry.feed_url, entry.id)
     assert 'no such entry' in excinfo.value.message
     assert reader.get_entry(entry_arg(entry.as_entry(feed=feed)), None) == None
-    assert reader.get_entry(entry_arg(entry.as_entry(feed=feed)), default=1) == 1
+    assert reader.get_entry(entry_arg(entry.as_entry(feed=feed)), 1) == 1
 
     reader._now = lambda: datetime(2010, 1, 3)
     reader.update_feeds()
@@ -1375,7 +1659,7 @@ def test_get_entry(reader, entry_arg):
     # TODO: find a way to not test added/last_updated here
 
     entry = entry.as_entry(
-        feed=feed._replace(
+        feed=feed.as_feed(
             added=datetime(2010, 1, 2), last_updated=datetime(2010, 1, 3)
         ),
         last_updated=datetime(2010, 1, 3),
@@ -1413,7 +1697,7 @@ def test_mark_as_important(reader, entry_arg):
 
     reader._storage = FakeStorage()
     entry = Entry('entry', None, feed=Feed('feed'))
-    reader.mark_as_important(entry_arg(entry))
+    reader.mark_entry_as_important(entry_arg(entry))
     assert reader._storage.calls == [
         ('mark_as_important_unimportant', 'feed', 'entry', True)
     ]
@@ -1422,7 +1706,7 @@ def test_mark_as_important(reader, entry_arg):
 def test_mark_as_unimportant(reader, entry_arg):
     reader._storage = FakeStorage()
     entry = Entry('entry', None, feed=Feed('feed'))
-    reader.mark_as_unimportant(entry_arg(entry))
+    reader.mark_entry_as_unimportant(entry_arg(entry))
     assert reader._storage.calls == [
         ('mark_as_important_unimportant', 'feed', 'entry', False)
     ]
@@ -1431,7 +1715,9 @@ def test_mark_as_unimportant(reader, entry_arg):
 @pytest.mark.parametrize(
     'exc', [EntryNotFoundError('feed', 'entry'), StorageError('whatever')]
 )
-@pytest.mark.parametrize('meth', ['mark_as_important', 'mark_as_unimportant'])
+@pytest.mark.parametrize(
+    'meth', ['mark_entry_as_important', 'mark_entry_as_unimportant']
+)
 def test_mark_as_important_unimportant_error(reader, exc, meth):
     reader._storage = FakeStorage(exc=exc)
     with pytest.raises(Exception) as excinfo:
@@ -1460,7 +1746,7 @@ def test_closed(reader):
 
 def test_direct_instantiation():
     with pytest.warns(UserWarning):
-        Reader('storage', 'search', 'parser')
+        Reader('storage', 'search', 'parser', DEFAULT_RESERVED_NAME_SCHEME)
 
 
 # BEGIN entry filtering tests
@@ -1544,8 +1830,8 @@ def test_entries_filtering(reader, pre_stuff, call_method, kwargs, expected):
     reader.add_feed(two.url)
     reader.update_feeds()
 
-    reader.mark_as_read((one.url, one_two.id))
-    reader.mark_as_important((one.url, one_three.id))
+    reader.mark_entry_as_read((one.url, one_two.id))
+    reader.mark_entry_as_important((one.url, one_three.id))
 
     pre_stuff(reader)
 
@@ -1593,10 +1879,10 @@ def test_entries_filtering_error(reader, pre_stuff, call_method, kwargs):
         (dict(feed_root='/path'), '/path'),
     ],
 )
-def test_make_reader_feed_root(monkeypatch, kwargs, feed_root):
+def test_make_reader_feed_root(monkeypatch, make_reader, kwargs, feed_root):
     exc = Exception("whatever")
 
-    def default_parser(feed_root):
+    def default_parser(feed_root, **kwargs):
         default_parser.feed_root = feed_root
         raise exc
 
@@ -1679,7 +1965,7 @@ def test_tags_basic(reader, chunk_size):
         'tag-common',
     ]
 
-    reader.remove_feed('two')
+    reader.delete_feed('two')
     assert list(reader.get_feed_tags('one')) == ['tag-1', 'tag-common']
     assert list(reader.get_feed_tags('two')) == []
     assert list(reader.get_feed_tags()) == ['tag-1', 'tag-common']
@@ -2001,7 +2287,7 @@ def test_change_feed_url_second_update(reader, new_feed_url):
     reader.enable_search()
     reader.update_search()
 
-    reader.remove_feed('2')
+    reader.delete_feed('2')
 
     old_one = reader.get_feed('1')
 
@@ -2062,7 +2348,7 @@ def test_change_feed_url_search_entry_id_repeats(reader):
 
     # TODO: write another test to make sure things marked as to_update remain marked after change
 
-    reader.remove_feed(two)
+    reader.delete_feed(two)
     reader.change_feed_url(one, two)
 
     new = set(reader.get_entries(feed=two))
@@ -2153,12 +2439,12 @@ def test_change_feed_url_search(reader):
 
 @rename_argument('reader', 'reader_with_two_feeds')
 def test_change_feed_url_metadata(reader):
-    reader.set_feed_metadata('1', 'key', 'value')
+    reader.set_feed_metadata_item('1', 'key', 'value')
 
     reader.change_feed_url('1', '3')
 
-    assert dict(reader.iter_feed_metadata('1')) == {}
-    assert dict(reader.iter_feed_metadata('3')) == {'key': 'value'}
+    assert dict(reader.get_feed_metadata('1')) == {}
+    assert dict(reader.get_feed_metadata('3')) == {'key': 'value'}
 
 
 @rename_argument('reader', 'reader_with_two_feeds')
@@ -2408,9 +2694,9 @@ def test_entry_counts(reader, kwargs, expected, pre_stuff, call_method, rv_type)
     pre_stuff(reader)
 
     for entry in two_entries[:2]:
-        reader.mark_as_read(entry)
+        reader.mark_entry_as_read(entry)
     for entry in two_entries[:4]:
-        reader.mark_as_important(entry)
+        reader.mark_entry_as_important(entry)
 
     rv = call_method(reader, **kwargs)
     assert type(rv) is rv_type
@@ -2556,3 +2842,74 @@ def test_logging_defaults():
     assert logger.level == logging.NOTSET
     assert len(logger.handlers) == 1
     assert isinstance(logger.handlers[0], logging.NullHandler)
+
+
+@pytest.mark.parametrize(
+    'kwargs, expected_timeout',
+    [
+        ({}, reader._parser.SESSION_TIMEOUT),
+        ({'session_timeout': (1.234, 324.1)}, (1.234, 324.1)),
+    ],
+)
+def test_session_timeout(monkeypatch, make_reader, kwargs, expected_timeout):
+    def send(*args, timeout=None, **kwargs):
+        raise Exception('timeout', timeout)
+
+    monkeypatch.setattr('requests.adapters.HTTPAdapter.send', send)
+
+    reader = make_reader(':memory:', **kwargs)
+    reader.add_feed('http://www.example.com')
+
+    with pytest.raises(ParseError) as exc_info:
+        reader.update_feed('http://www.example.com')
+
+    assert exc_info.value.__cause__.args == ('timeout', expected_timeout)
+
+
+def test_plugins(monkeypatch, make_reader):
+    def one(reader):
+        one.reader = reader
+
+    def two(reader):
+        two.reader = reader
+
+    from reader.plugins import _PLUGINS
+
+    monkeypatch.setitem(_PLUGINS, 'one', one)
+
+    reader = make_reader(':memory:', plugins=['one', two])
+
+    assert one.reader is reader
+    assert two.reader is reader
+
+    with pytest.raises(InvalidPluginError) as exc_info:
+        make_reader(':memory:', plugins=['two'])
+
+    assert str(exc_info.value) == "no such built-in plugin: 'two'"
+
+
+def test_reserved_names(make_reader):
+    with pytest.raises(ValueError):
+        make_reader(':memory:', reserved_name_scheme={})
+
+    reader = make_reader(':memory:')
+
+    assert reader.make_reader_reserved_name('key') == '.reader.key'
+    assert reader.make_plugin_reserved_name('myplugin') == '.plugin.myplugin'
+    assert reader.make_plugin_reserved_name('myplugin', 'key') == '.plugin.myplugin.key'
+
+    with pytest.raises(AttributeError):
+        reader.reserved_name_scheme = {}
+
+    new_scheme = {'reader_prefix': '', 'plugin_prefix': '.', 'separator': ':'}
+
+    reader.reserved_name_scheme = new_scheme
+
+    assert reader.make_reader_reserved_name('key') == 'key'
+    assert reader.make_plugin_reserved_name('myplugin') == '.myplugin'
+    assert reader.make_plugin_reserved_name('myplugin', 'key') == '.myplugin:key'
+
+    assert dict(reader.reserved_name_scheme) == new_scheme
+
+    with pytest.raises(TypeError):
+        reader.reserved_name_scheme['separator'] = '.'
