@@ -11,7 +11,8 @@ from reader import make_reader
 from reader._storage import Storage
 
 
-def test_update_stale(reader, call_update_method):
+@pytest.mark.parametrize('entry_updated', [datetime(2010, 1, 1), None])
+def test_update_stale(reader, call_update_method, entry_updated):
     """When a feed is marked as stale feeds/entries should be updated
     regardless of their .updated or caching headers.
 
@@ -22,7 +23,7 @@ def test_update_stale(reader, call_update_method):
     reader._parser = parser
 
     feed = parser.feed(1, datetime(2010, 1, 1))
-    entry = parser.entry(1, 1, datetime(2010, 1, 1))
+    entry = parser.entry(1, 1, entry_updated)
 
     with pytest.raises(FeedNotFoundError):
         reader._storage.mark_as_stale(feed.url)
@@ -35,24 +36,29 @@ def test_update_stale(reader, call_update_method):
     assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
         (feed.url, feed.title, datetime(2010, 1, 1))
     }
-    # FIXME: use entry.last_updated once we have it
-    assert set((e.id, e.title) for e in reader.get_entries()) == {
-        (entry.id, entry.title)
+    assert set((e.id, e.title, e.last_updated) for e in reader.get_entries()) == {
+        (entry.id, entry.title, datetime(2010, 1, 1))
     }
 
-    new_feed = parser.feed(1, datetime(2010, 1, 1), title="new feed title")
-    new_entry = parser.entry(1, 1, datetime(2010, 1, 1), title="new entry title")
+    # we can't change feed/entry here because their hash would change,
+    # resulting in an update;
+    # the only way to check they were updated is through last_updated
 
-    # nothing changes after update
-    reader._now = lambda: datetime(2010, 1, 2)
-    call_update_method(reader, feed.url)
-    assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
-        (feed.url, feed.title, datetime(2010, 1, 1))
-    }
-    # FIXME: use entry.last_updated once we have it
-    assert set((e.id, e.title) for e in reader.get_entries()) == {
-        (entry.id, entry.title)
-    }
+    # should we deprecate the staleness API? maybe:
+    # https://github.com/lemon24/reader/issues/179#issuecomment-663840297
+    # OTOH, we may still want an update to happen for other side-effects,
+    # even if the hash doesn't change
+
+    if entry_updated:
+        # nothing changes after update
+        reader._now = lambda: datetime(2010, 1, 2)
+        call_update_method(reader, feed.url)
+        assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
+            (feed.url, feed.title, datetime(2010, 1, 1))
+        }
+        assert set((e.id, e.title, e.last_updated) for e in reader.get_entries()) == {
+            (entry.id, entry.title, datetime(2010, 1, 1))
+        }
 
     # but it does if we mark the feed as stale
     parser.calls[:] = []
@@ -61,11 +67,10 @@ def test_update_stale(reader, call_update_method):
     call_update_method(reader, feed.url)
     assert parser.calls == [(feed.url, None, None)]
     assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
-        (feed.url, new_feed.title, datetime(2010, 1, 3))
+        (feed.url, feed.title, datetime(2010, 1, 3))
     }
-    # FIXME: use entry.last_updated once we have it
-    assert set((e.id, e.title) for e in reader.get_entries()) == {
-        (entry.id, new_entry.title)
+    assert set((e.id, e.title, e.last_updated) for e in reader.get_entries()) == {
+        (entry.id, entry.title, datetime(2010, 1, 3))
     }
 
 
@@ -128,7 +133,34 @@ def test_post_entry_add_plugins(reader):
     # TODO: What is the expected behavior if a plugin raises an exception?
 
 
-def test_make_reader_storage():
-    storage = Storage(':memory:')
+def test_post_feed_update_plugins(reader):
+    parser = Parser()
+    reader._parser = parser
+
+    plugin_calls = []
+
+    def first_plugin(r, e):
+        assert r is reader
+        plugin_calls.append((first_plugin, e))
+
+    def second_plugin(r, e):
+        assert r is reader
+        plugin_calls.append((second_plugin, e))
+
+    feed = parser.feed(1, datetime(2010, 1, 1))
+    parser.entry(1, 1, datetime(2010, 1, 1))
+    reader.add_feed(feed.url)
+    reader._post_feed_update_plugins.append(first_plugin)
+    reader._post_feed_update_plugins.append(second_plugin)
+
+    reader.update_feeds()
+    assert plugin_calls == [
+        (first_plugin, feed.url),
+        (second_plugin, feed.url),
+    ]
+    assert set(e.id for e in reader.get_entries()) == {'1, 1'}
+
+
+def test_make_reader_storage(storage):
     reader = make_reader('', _storage=storage)
     assert reader._storage is storage
