@@ -1,11 +1,17 @@
 import sqlite3
 import sys
+from functools import wraps
 
 import pytest
+from utils import rename_argument
 
+from reader._sqlite_utils import DBError
 from reader._sqlite_utils import ddl_transaction
 from reader._sqlite_utils import HeavyMigration
+from reader._sqlite_utils import IdError
 from reader._sqlite_utils import IntegrityError
+from reader._sqlite_utils import NewMigration
+from reader._sqlite_utils import OldMigration
 from reader._sqlite_utils import require_compile_options
 from reader._sqlite_utils import require_version
 from reader._sqlite_utils import RequirementError
@@ -14,6 +20,20 @@ from reader._sqlite_utils import setup_db
 from reader._sqlite_utils import wrap_exceptions
 
 
+original_sqlite3_connect = sqlite3.connect
+
+
+@pytest.fixture(autouse=True)
+def patch_sqlite3_connect(monkeypatch, request):
+    @wraps(original_sqlite3_connect)
+    def connect(*args, **kwargs):
+        db = original_sqlite3_connect(*args, **kwargs)
+        request.addfinalizer(db.close)
+        return db
+
+    monkeypatch.setattr('sqlite3.connect', connect)
+
+
 def dummy_ddl_transaction(db):
     """Just use a regular transaction."""
     return db
@@ -131,18 +151,29 @@ def update_from_1_to_2_error(db):
     raise WeirdError('update')
 
 
-def test_migration_create():
+@pytest.fixture(params=[HeavyMigration, NewMigration])
+def migration_cls_no_id(request):
+    # TODO: delete in 2.0; needed for the migration from OldMigration to NewMigration
+    return request.param
+
+
+@pytest.fixture(params=[0, 1234])
+def migration_cls(request, migration_cls_no_id):
+    return lambda *a, **kw: migration_cls_no_id(*a, id=request.param, **kw)
+
+
+def test_migration_create(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_2, 2, {})
+    migration = migration_cls(create_db_2, 2, {})
     # should call migration.create but not migration.migrations[1]
     migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one', 'two']
 
 
-def test_migration_create_error():
+def test_migration_create_error(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_2_error, 2, {})
+    migration = migration_cls(create_db_2_error, 2, {})
     # should call migration.create but not migration.migrations[1]
     with pytest.raises(WeirdError) as excinfo:
         migration.migrate(db)
@@ -151,33 +182,33 @@ def test_migration_create_error():
     assert columns == []
 
 
-def test_migration_update():
+def test_migration_update(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_1, 1, {})
+    migration = migration_cls(create_db_1, 1, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_2_error, 2, {1: update_from_1_to_2})
+    migration = migration_cls(create_db_2_error, 2, {1: update_from_1_to_2})
     # should call migration.migrations[1] but not migration.create
     migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one', 'two']
 
 
-def test_migration_no_update():
+def test_migration_no_update(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_2, 2, {})
+    migration = migration_cls(create_db_2, 2, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_2_error, 2, {})
+    migration = migration_cls(create_db_2_error, 2, {})
     # should call neither migration.create nor migration.migrations[1]
     migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one', 'two']
 
 
-def test_migration_update_error():
+def test_migration_update_error(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_1, 1, {})
+    migration = migration_cls(create_db_1, 1, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_2_error, 2, {1: update_from_1_to_2_error})
+    migration = migration_cls(create_db_2_error, 2, {1: update_from_1_to_2_error})
     # should call migration.migrations[1] but not migration.create
     with pytest.raises(WeirdError) as excinfo:
         migration.migrate(db)
@@ -186,49 +217,40 @@ def test_migration_update_error():
     assert columns == ['one']
 
 
-def test_migration_unsupported_old_version():
+def test_migration_unsupported_old_version(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_1, 1, {})
+    migration = migration_cls(create_db_1, 1, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_2, 2, {})
+    migration = migration_cls(create_db_2, 2, {})
     with pytest.raises(SchemaVersionError) as excinfo:
         migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one']
 
 
-def test_migration_unsupported_intermediary_version():
+def test_migration_unsupported_intermediary_version(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_1, 1, {})
+    migration = migration_cls(create_db_1, 1, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_2, 3, {1: update_from_1_to_2})
+    migration = migration_cls(create_db_2, 3, {1: update_from_1_to_2})
     with pytest.raises(SchemaVersionError) as excinfo:
         migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one']
 
 
-def test_migration_invalid_version():
+def test_migration_invalid_version(migration_cls):
     db = sqlite3.connect(':memory:')
-    migration = HeavyMigration(create_db_2, 2, {})
+    migration = migration_cls(create_db_2, 2, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db_1, 1, {})
+    migration = migration_cls(create_db_1, 1, {})
     with pytest.raises(SchemaVersionError) as excinfo:
         migration.migrate(db)
     columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
     assert columns == ['one', 'two']
 
 
-def test_migration_invalid_version_table():
-    db = sqlite3.connect(':memory:')
-    with db:
-        db.execute("CREATE TABLE version (not_version);")
-    migration = HeavyMigration(create_db_2, 2, {})
-    with pytest.raises(SchemaVersionError) as excinfo:
-        migration.migrate(db)
-
-
-def test_migration_integrity_error():
+def test_migration_integrity_error(migration_cls):
     def create_db(db):
         db.execute("CREATE TABLE t (one INTEGER PRIMARY KEY);")
         db.execute(
@@ -241,13 +263,101 @@ def test_migration_integrity_error():
     db = sqlite3.connect(':memory:')
     db.execute("PRAGMA foreign_keys = ON;")
 
-    migration = HeavyMigration(create_db, 1, {})
+    migration = migration_cls(create_db, 1, {})
     migration.migrate(db)
-    migration = HeavyMigration(create_db, 2, {1: update_from_1_to_2})
+    migration = migration_cls(create_db, 2, {1: update_from_1_to_2})
     with pytest.raises(IntegrityError):
         migration.migrate(db)
 
 
+@pytest.mark.parametrize('version', [-1, []])
+def test_migration_version_valuerror(migration_cls, version):
+    db = sqlite3.connect(':memory:')
+    migration = migration_cls(create_db_1, version, {})
+    with pytest.raises(ValueError) as excinfo:
+        migration.migrate(db)
+
+
+def test_old_migration_invalid_version_table():
+    # TODO: remove in 2.0
+    db = sqlite3.connect(':memory:')
+    with db:
+        db.execute("CREATE TABLE version (not_version);")
+    migration = OldMigration(create_db_2, 2, {})
+    with pytest.raises(SchemaVersionError) as excinfo:
+        migration.migrate(db)
+
+
+def test_migration_version_migration():
+    """That is, the migration of the migration system from storing
+    the schema version in a "version (version INTEGER NOT NULL)" table
+    to using "PRAGMA user_version".
+    """
+    # TODO: remove in 2.0
+    db = sqlite3.connect(':memory:')
+    with db:
+        # emulate a database managed by OldMigration
+        db.execute("CREATE TABLE version (version INTEGER NOT NULL);")
+        db.execute("INSERT INTO version VALUES (1);")
+        create_db_1(db)
+    migration = HeavyMigration(create_db_1, 2, {1: update_from_1_to_2})
+    migration.migrate(db)
+
+    with pytest.raises(sqlite3.OperationalError) as excinfo:
+        db.execute("SELECT * FROM version;")
+    assert 'no such table' in str(excinfo.value)
+
+    assert list(db.execute("PRAGMA user_version;")) == [(2,)]
+
+    columns = [r[1] for r in db.execute("PRAGMA table_info(t);")]
+    assert columns == ['one', 'two']
+
+
+def test_nonempty(migration_cls):
+    db = sqlite3.connect(':memory:')
+    db.execute("CREATE TABLE unexpected (one INTEGER);")
+    migration = migration_cls(create_db_2, 2, {})
+    with pytest.raises(DBError):
+        migration.migrate(db)
+
+
+@rename_argument('migration_cls', 'migration_cls_no_id')
+def test_invalid_id(migration_cls):
+    db = sqlite3.connect(':memory:')
+    db.execute("PRAGMA application_id = 2;")
+    migration = migration_cls(create_db_2, 2, {}, id=1)
+    with pytest.raises(IdError):
+        migration.migrate(db)
+
+
+@rename_argument('migration_cls', 'migration_cls_no_id')
+def test_missing_id_with_version(migration_cls):
+    db = sqlite3.connect(':memory:')
+    migration_cls.set_version(db, 2)
+    migration = migration_cls(create_db_2, 2, {}, id=1)
+    with pytest.raises(IdError):
+        migration.migrate(db)
+
+
+@rename_argument('migration_cls', 'migration_cls_no_id')
+def test_missing_id_after_migration(migration_cls):
+    db = sqlite3.connect(':memory:')
+
+    migration = migration_cls(create_db_1, 1, {})
+    migration.migrate(db)
+
+    migration = migration_cls(create_db_2_error, 2, {1: update_from_1_to_2}, id=1)
+    with pytest.raises(IdError):
+        migration.migrate(db)
+
+    def bad_update_from_1_to_2(db):
+        migration_cls.set_id(db, 2)
+
+    migration = migration_cls(create_db_2_error, 2, {1: bad_update_from_1_to_2}, id=1)
+    with pytest.raises(IdError):
+        migration.migrate(db)
+
+
 def test_require_version():
     db = MockConnection(execute_rv=[('3.15.0',)])
 
@@ -296,8 +406,9 @@ def test_setup_db_wal_enabled(db_path, wal_enabled, expected_mode):
     setup_db(
         db,
         create=lambda db: None,
-        version=0,
+        version=1,
         migrations={},
+        id=1234,
         minimum_sqlite_version=(3, 15, 0),
         wal_enabled=wal_enabled,
     )
