@@ -105,8 +105,8 @@ def add_feed(storage, feed, __):
     storage.add_feed(feed.url + '_', datetime(2010, 1, 1))
 
 
-def remove_feed(storage, feed, __):
-    storage.remove_feed(feed.url)
+def delete_feed(storage, feed, __):
+    storage.delete_feed(feed.url)
 
 
 def get_feeds(storage, _, __):
@@ -147,13 +147,13 @@ def update_feed_last_updated(storage, feed, entry):
 
 def add_or_update_entry(storage, feed, entry):
     storage.add_or_update_entry(
-        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0)
+        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0, 0)
     )
 
 
 def add_or_update_entries(storage, feed, entry):
     storage.add_or_update_entries(
-        [EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0)]
+        [EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0, 0)]
     )
 
 
@@ -165,28 +165,28 @@ def get_entries_chunk_size_1(storage, _, __):
     list(storage.get_entries_page(chunk_size=1, now=datetime(2010, 1, 1)))
 
 
-def iter_feed_metadata(storage, feed, __):
-    list(storage.iter_feed_metadata(feed.url))
+def iter_metadata(storage, feed, __):
+    list(storage.iter_metadata((feed.url,)))
 
 
-def set_feed_metadata(storage, feed, __):
-    storage.set_feed_metadata(feed.url, 'key', 'value')
+def set_metadata(storage, feed, __):
+    storage.set_metadata((feed.url,), 'key', 'value')
 
 
-def delete_feed_metadata(storage, feed, __):
-    storage.delete_feed_metadata(feed.url, 'key')
+def delete_metadata(storage, feed, __):
+    storage.delete_metadata((feed.url,), 'key')
 
 
-def add_feed_tag(storage, feed, __):
-    storage.add_feed_tag(feed.url, 'tag')
+def add_tag(storage, feed, __):
+    storage.add_tag((feed.url,), 'tag')
 
 
-def remove_feed_tag(storage, feed, __):
-    storage.remove_feed_tag(feed.url, 'tag')
+def remove_tag(storage, feed, __):
+    storage.remove_tag((feed.url,), 'tag')
 
 
-def get_feed_tags(storage, feed, __):
-    list(storage.get_feed_tags(feed.url))
+def get_tags(storage, feed, __):
+    list(storage.get_tags((feed.url,)))
 
 
 def get_feed_counts(storage, _, __):
@@ -211,7 +211,7 @@ def get_entry_last(storage, feed, entry):
     [
         init,
         add_feed,
-        remove_feed,
+        delete_feed,
         get_feeds,
         get_feeds_for_update,
         get_entries_for_update,
@@ -225,12 +225,12 @@ def get_entry_last(storage, feed, entry):
         add_or_update_entries,
         get_entries_chunk_size_0,
         get_entries_chunk_size_1,
-        iter_feed_metadata,
-        set_feed_metadata,
-        delete_feed_metadata,
-        add_feed_tag,
-        remove_feed_tag,
-        get_feed_tags,
+        iter_metadata,
+        set_metadata,
+        delete_metadata,
+        add_tag,
+        remove_tag,
+        get_tags,
         get_feed_counts,
         get_entry_counts,
         get_feed_last,
@@ -254,7 +254,7 @@ def check_errors_locked(db_path, pre_stuff, do_stuff, exc_type):
     entry = EntryData('one', 'entry', datetime(2010, 1, 2))
     storage.add_feed(feed.url, datetime(2010, 1, 1))
     storage.add_or_update_entry(
-        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0)
+        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0, 0)
     )
 
     in_transaction = threading.Event()
@@ -309,12 +309,12 @@ def iter_pagination_chunk_size_3(storage):
     return storage.get_entries_page(chunk_size=3, now=datetime(2010, 1, 1))
 
 
-def iter_iter_feed_metadata(storage):
-    return storage.iter_feed_metadata_page('two', chunk_size=1)
+def iter_iter_metadata(storage):
+    return storage.iter_metadata_page(('two',), chunk_size=1)
 
 
-def iter_get_feed_tags(storage):
-    return storage.get_feed_tags('two')
+def iter_get_tags(storage):
+    return storage.get_tags_page(('two',), chunk_size=1)
 
 
 @pytest.mark.slow
@@ -330,8 +330,8 @@ def iter_get_feed_tags(storage):
         iter_pagination_chunk_size_1,
         iter_pagination_chunk_size_2,
         iter_pagination_chunk_size_3,
-        iter_iter_feed_metadata,
-        iter_get_feed_tags,
+        iter_iter_metadata,
+        iter_get_tags,
     ],
 )
 def test_iter_locked(db_path, iter_stuff):
@@ -350,18 +350,18 @@ def check_iter_locked(db_path, pre_stuff, iter_stuff):
     entry = EntryData('one', 'entry', datetime(2010, 1, 1), title='entry')
     storage.add_feed(feed.url, datetime(2010, 1, 2))
     storage.add_or_update_entry(
-        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0)
+        EntryUpdateIntent(entry, entry.updated, datetime(2010, 1, 1), 0, 0)
     )
     storage.add_feed('two', datetime(2010, 1, 1))
     storage.add_or_update_entry(
         EntryUpdateIntent(
-            entry._replace(feed_url='two'), entry.updated, datetime(2010, 1, 1), 0
+            entry._replace(feed_url='two'), entry.updated, datetime(2010, 1, 1), 0, 0
         )
     )
-    storage.set_feed_metadata('two', '1', 1)
-    storage.set_feed_metadata('two', '2', 2)
-    storage.add_feed_tag('two', '1')
-    storage.add_feed_tag('two', '2')
+    storage.set_metadata(('two',), '1', 1)
+    storage.set_metadata(('two',), '2', 2)
+    storage.add_tag(('two',), '1')
+    storage.add_tag(('two',), '2')
 
     if pre_stuff:
         pre_stuff(storage)
@@ -382,6 +382,8 @@ def test_update_feed_last_updated_not_found(db_path):
         storage.update_feed(FeedUpdateIntent('inexistent-feed', datetime(2010, 1, 2)))
 
 
+# aviod spurious PyPy "sqlite3.InterfaceError: Error binding parameter X")
+@pytest.mark.skipif("sys.implementation.name == 'pypy'")
 @pytest.mark.parametrize(
     'entry_count',
     [
@@ -392,16 +394,7 @@ def test_update_feed_last_updated_not_found(db_path):
         # variable_number defaults to 250000 in Ubuntu 18.04 -provided SQLite
         pytest.param(
             int(250000 / 2) + 1,
-            marks=(
-                pytest.mark.slow,
-                # Fails on PyPy3 7.3.1 *only when running with --cov* with:
-                #   _sqlite3.InterfaceError: Error binding parameter :feed_url - probably unsupported type.
-                # but not CPython or PyPy3 7.2.0 (on macOS, at least).
-                pytest.mark.xfail(
-                    "sys.implementation.name == 'pypy' "
-                    "and sys.pypy_version_info[:2] == (7, 3)",
-                ),
-            ),
+            marks=(pytest.mark.slow,),
         ),
     ],
 )
@@ -444,17 +437,19 @@ class StorageAlwaysGetEntriesForUpdateFallback(Storage):
 def test_get_entries_for_update(storage_cls):
     storage = storage_cls(':memory:')
     storage.add_feed('feed', datetime(2010, 1, 1))
+    entry = EntryData('feed', 'one', datetime(2010, 1, 1))
     storage.add_or_update_entry(
         EntryUpdateIntent(
-            EntryData('feed', 'one', datetime(2010, 1, 1)),
+            entry,
             datetime(2010, 1, 2),
             datetime(2010, 1, 1),
             0,
+            0,
         )
     )
 
     assert list(storage.get_entries_for_update([('feed', 'one'), ('feed', 'two')])) == [
-        EntryForUpdate(datetime(2010, 1, 1)),
+        EntryForUpdate(datetime(2010, 1, 1), entry.hash, 0),
         None,
     ]
 
@@ -474,6 +469,7 @@ def test_entry_remains_read_after_update(storage_with_two_entries):
             datetime(2010, 1, 2),
             datetime(2010, 1, 2),
             0,
+            0,
         )
     )
 
@@ -494,6 +490,7 @@ def storage_with_two_entries(storage):
             datetime(2010, 1, 2),
             datetime(2010, 1, 2),
             0,
+            0,
         )
     )
     storage.add_or_update_entry(
@@ -502,6 +499,7 @@ def storage_with_two_entries(storage):
             datetime(2010, 1, 2),
             datetime(2010, 1, 2),
             1,
+            0,
         )
     )
     return storage
@@ -555,6 +553,7 @@ def test_important_entry_remains_important_after_update(storage):
             datetime(2010, 1, 2),
             datetime(2010, 1, 2),
             0,
+            0,
         )
     )
 
